@@ -58,8 +58,6 @@
 #include <osmocom/bsc/bss.h>
 #include <osmocom/bsc/bts.h>
 
-struct gsm_network *bsc_gsmnet;
-
 static int net_listen_testnr;
 static int restart;
 static bool get_attr;
@@ -144,21 +142,22 @@ static int ia_config_connect(struct gsm_bts *bts, struct sockaddr_in *sin)
 	}
 	line->ops = &ipaccess_e1inp_line_ops;
 
-	/* create E1 timeslots for signalling and TRAU frames */
-	e1inp_ts_config_sign(&line->ts[1-1], line);
-	e1inp_ts_config_sign(&line->ts[2-1], line);
+	sign_ts = e1inp_line_ipa_oml_ts(line);
+	rsl_ts = e1inp_line_ipa_rsl_ts(line, 0);
 
-	/* create signalling links for TS1 */
-	sign_ts = &line->ts[1-1];
-	rsl_ts = &line->ts[2-1];
+	/* create E1 timeslots for signalling and TRAU frames */
+	e1inp_ts_config_sign(sign_ts, line);
+	e1inp_ts_config_sign(rsl_ts, line);
+
+	/* create signalling links for TRX0 */
 	oml_link = e1inp_sign_link_create(sign_ts, E1INP_SIGN_OML,
-					  bts->c0, 0xff, 0);
+					  bts->c0, IPAC_PROTO_OML, 0);
 	rsl_link = e1inp_sign_link_create(rsl_ts, E1INP_SIGN_RSL,
-					  bts->c0, 0, 0);
+					  bts->c0, IPAC_PROTO_RSL, 0);
 
 	/* create back-links from bts/trx */
 	bts->oml_link = oml_link;
-	bts->c0->rsl_link = rsl_link;
+	bts->c0->rsl_link_primary = rsl_link;
 
 	/* default port at BTS for incoming connections is 3006 */
 	if (sin->sin_port == 0)
@@ -502,7 +501,7 @@ static const struct value_string ipa_nvflag_strs[] = {
 	{ 0x0002, "static-gw" },
 	{ 0x0004, "no-dhcp-vsi" },
 	{ 0x0008, "dhcp-enabled" },
-	{ 0x0040, "led-disabled" },
+	{ 0x0040, "led-enabled" },
 	{ 0x0100, "secondary-oml-enabled" },
 	{ 0x0200, "diag-enabled" },
 	{ 0x0400, "cli-enabled" },
@@ -517,8 +516,10 @@ static int ipa_nvflag_set(uint16_t *flags, uint16_t *mask, const char *name, int
 {
 	int rc;
 	rc = get_string_value(ipa_nvflag_strs, name);
-	if (rc < 0)
+	if (rc < 0) {
+		fprintf(stderr, "Unknown attribute '%s'\n", name);
 		return rc;
+	}
 
 	*mask |= rc;
 	if (en)
@@ -926,7 +927,7 @@ static const struct log_info_cat log_categories[] = {
 		.name = "DNM",
 		.description = "A-bis Network Management / O&M (NM/OML)",
 		.color = "\033[1;36m",
-		.loglevel = LOGL_DEBUG,
+		.loglevel = LOGL_NOTICE,
 		.enabled = 1,
 	},
 };
@@ -1121,7 +1122,7 @@ int main(int argc, char **argv)
 	}
 
 	bts->oml_link->ts->sign.delay = 10;
-	bts->c0->rsl_link->ts->sign.delay = 10;
+	bts->c0->rsl_link_primary->ts->sign.delay = 10;
 	while (1) {
 		rc = osmo_select_main(0);
 		if (rc < 0)
@@ -1130,15 +1131,3 @@ int main(int argc, char **argv)
 
 	exit(0);
 }
-
-/* Stub */
-int osmo_bsc_sigtran_send(struct gsm_subscriber_connection *conn, struct msgb *msg)
-{ return 0; }
-int osmo_bsc_sigtran_open_conn(struct gsm_subscriber_connection *conn, struct msgb *msg)
-{ return 0; }
-int rsl_chan_ms_power_ctrl(struct gsm_lchan *lchan) { return 0; }
-void pcu_info_update(struct gsm_bts *bts) {};
-int rsl_sacch_filling(struct gsm_bts_trx *trx, uint8_t type, const uint8_t *data, int len) { return 0; }
-int rsl_bcch_info(const struct gsm_bts_trx *trx, enum osmo_sysinfo_type si_type, const uint8_t *data, int len)
-{ return 0; }
-int gsm_generate_si(struct gsm_bts *bts, enum osmo_sysinfo_type si_type) { return 0; }
