@@ -45,6 +45,7 @@
 #include <osmocom/bsc/assignment_fsm.h>
 #include <osmocom/bsc/codec_pref.h>
 #include <osmocom/mgcp_client/mgcp_client_endpoint_fsm.h>
+#include <osmocom/mgcp_client/mgcp_client_pool.h>
 #include <osmocom/core/byteswap.h>
 #include <osmocom/bsc/lb.h>
 #include <osmocom/bsc/lcs_loc_req.h>
@@ -161,7 +162,7 @@ static void gscon_bssmap_clear(struct gsm_subscriber_connection *conn,
 		return;
 	}
 
-	rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_CLEAR_RQST]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_CLEAR_RQST));
 	rc = osmo_bsc_sigtran_send(conn, resp);
 	if (rc < 0)
 		LOGPFSML(conn->fi, LOGL_ERROR, "Unable to deliver BSSMAP Clear Request message\n");
@@ -176,7 +177,7 @@ static void forward_dtap(struct gsm_subscriber_connection *conn, struct msgb *ms
 	OSMO_ASSERT(conn);
 
 	resp = gsm0808_create_dtap(msg, OBSC_LINKID_CB(msg));
-	rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_DTAP]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_DTAP));
 	gscon_sigtran_send(conn, resp);
 }
 
@@ -195,7 +196,8 @@ static void gscon_release_lchan(struct gsm_subscriber_connection *conn, struct g
 		conn->ho.new_lchan = NULL;
 	if (conn->assignment.new_lchan == lchan)
 		conn->assignment.new_lchan = NULL;
-	lchan_release(lchan, do_rr_release, err, cause_rr);
+	lchan_release(lchan, do_rr_release, err, cause_rr,
+		      gscon_last_eutran_plmn(conn));
 }
 
 void gscon_release_lchans(struct gsm_subscriber_connection *conn, bool do_rr_release, enum gsm48_rr_cause cause_rr)
@@ -517,39 +519,47 @@ struct osmo_mgcpc_ep *gscon_ensure_mgw_endpoint(struct gsm_subscriber_connection
 						uint16_t msc_assigned_cic, struct gsm_lchan *for_lchan)
 {
 	const char *epname;
+	struct mgcp_client *mgcp_client = NULL;
 
 	if (conn->user_plane.mgw_endpoint)
 		return conn->user_plane.mgw_endpoint;
 
+	if (gscon_is_sccplite(conn) || gscon_is_aoip(conn)) {
+		/* Get MGCP client from pool */
+		mgcp_client = mgcp_client_pool_get(conn->network->mgw.mgw_pool);
+		if (!mgcp_client) {
+			LOGPFSML(conn->fi, LOGL_ERROR,
+				 "cannot ensure MGW endpoint -- no MGW configured, check configuration!\n");
+			conn->user_plane.mgw_endpoint = NULL;
+			return NULL;
+		}
+	}
+
 	if (gscon_is_sccplite(conn)) {
 		/* derive endpoint name from CIC on A interface side */
 		conn->user_plane.mgw_endpoint =
 			osmo_mgcpc_ep_alloc(conn->fi, GSCON_EV_FORGET_MGW_ENDPOINT,
-					    conn->network->mgw.client,
+					    mgcp_client,
 					    conn->network->mgw.tdefs,
 					    conn->fi->id,
 					    "%x@%s", msc_assigned_cic,
-					    mgcp_client_endpoint_domain(conn->network->mgw.client));
+					    mgcp_client_endpoint_domain(mgcp_client));
 		LOGPFSML(conn->fi, LOGL_DEBUG, "MGW endpoint name derived from CIC 0x%x: %s\n",
 			 msc_assigned_cic, osmo_mgcpc_ep_name(conn->user_plane.mgw_endpoint));
 
 	} else if (gscon_is_aoip(conn)) {
-
 		if (is_ipaccess_bts(for_lchan->ts->trx->bts))
 			/* use dynamic RTPBRIDGE endpoint allocation in MGW */
-			epname = mgcp_client_rtpbridge_wildcard(conn->network->mgw.client);
+			epname = mgcp_client_rtpbridge_wildcard(mgcp_client);
 		else {
-			epname = mgcp_client_e1_epname(conn, conn->network->mgw.client, for_lchan->ts->e1_link.e1_nr,
+			epname = mgcp_client_e1_epname(conn, mgcp_client, for_lchan->ts->e1_link.e1_nr,
 						       for_lchan->ts->e1_link.e1_ts, 16,
 						       for_lchan->ts->e1_link.e1_ts_ss*2);
 		}
 
 		conn->user_plane.mgw_endpoint =
-			osmo_mgcpc_ep_alloc(conn->fi, GSCON_EV_FORGET_MGW_ENDPOINT,
-					    conn->network->mgw.client,
-					    conn->network->mgw.tdefs,
-					    conn->fi->id,
-					    "%s", epname);
+			osmo_mgcpc_ep_alloc(conn->fi, GSCON_EV_FORGET_MGW_ENDPOINT, mgcp_client,
+					    conn->network->mgw.tdefs, conn->fi->id, "%s", epname);
 	} else {
 		LOGPFSML(conn->fi, LOGL_ERROR, "Conn is neither SCCPlite nor AoIP!?\n");
 		return NULL;
@@ -599,7 +609,7 @@ bool gscon_connect_mgw_to_msc(struct gsm_subscriber_connection *conn,
 
 	ci = conn->user_plane.mgw_endpoint_ci_msc;
 	if (ci) {
-		const struct mgcp_conn_peer *prev_crcx_info = osmo_mgcpc_ep_ci_get_rtp_info(ci);
+		const struct mgcp_conn_peer *prev_crcx_info = osmo_mgcpc_ep_ci_get_remote_rtp_info(ci);
 
 		if (!conn->user_plane.mgw_endpoint) {
 			LOGPFSML(conn->fi, LOGL_ERROR, "Internal error: conn has a CI but no endpoint\n");
@@ -617,6 +627,8 @@ bool gscon_connect_mgw_to_msc(struct gsm_subscriber_connection *conn,
 			LOGPFSML(conn->fi, LOGL_DEBUG,
 				 "MSC side MGW endpoint ci is already configured to %s\n",
 				 osmo_mgcpc_ep_ci_name(ci));
+			/* Immediately dispatch the success event */
+			osmo_fsm_inst_dispatch(notify, event_success, notify_data);
 			return true;
 		}
 
@@ -696,9 +708,19 @@ void gscon_change_primary_lchan(struct gsm_subscriber_connection *conn, struct g
 	/* On release, do not receive release events that look like the primary lchan is gone. */
 	struct gsm_lchan *old_lchan = conn->lchan;
 
+	OSMO_ASSERT(new_lchan);
+
 	if (old_lchan == new_lchan)
 		return;
 
+	if (!old_lchan)
+		LOGPFSML(conn->fi, LOGL_DEBUG, "setting primary lchan for this conn to %s\n",
+			 new_lchan->fi? osmo_fsm_inst_name(new_lchan->fi) : gsm_lchan_name(new_lchan));
+	else
+		LOGPFSML(conn->fi, LOGL_DEBUG, "primary lchan for this conn changes from %s to %s\n",
+			 old_lchan->fi? osmo_fsm_inst_name(old_lchan->fi) : gsm_lchan_name(old_lchan),
+			 new_lchan->fi? osmo_fsm_inst_name(new_lchan->fi) : gsm_lchan_name(new_lchan));
+
 	conn->lchan = new_lchan;
 	conn->lchan->conn = conn;
 
@@ -788,6 +810,12 @@ void gscon_forget_lchan(struct gsm_subscriber_connection *conn, struct gsm_lchan
 
 static void gscon_forget_mgw_endpoint(struct gsm_subscriber_connection *conn)
 {
+	struct mgcp_client *mgcp_client;
+
+	/* Put MGCP client back into MGW pool */
+	mgcp_client = osmo_mgcpc_ep_client(conn->user_plane.mgw_endpoint);
+	mgcp_client_pool_put(mgcp_client);
+
 	conn->user_plane.mgw_endpoint = NULL;
 	conn->user_plane.mgw_endpoint_ci_msc = NULL;
 	conn->ho.created_ci_for_msc = NULL;
@@ -808,8 +836,9 @@ void gscon_forget_mgw_endpoint_ci(struct gsm_subscriber_connection *conn, struct
 static void gscon_fsm_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct gsm_subscriber_connection *conn = fi->priv;
-	const struct gscon_clear_cmd_data *ccd;
-	struct osmo_mobile_identity *mi_imsi;
+	const enum gsm0808_cause *cause_0808;
+	const struct tlv_parsed *tp;
+	struct osmo_mobile_identity mi_imsi;
 
 	/* Regular allstate event processing */
 	switch (event) {
@@ -825,14 +854,12 @@ static void gscon_fsm_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *d
 			osmo_fsm_inst_dispatch(conn->lcs.loc_req->fi, LCS_LOC_REQ_EV_CONN_CLEAR, NULL);
 
 		OSMO_ASSERT(data);
-		ccd = data;
-		if (conn->lchan)
-			conn->lchan->release.is_csfb = ccd->is_csfb;
+		cause_0808 = data;
 		/* MSC tells us to cleanly shut down */
 		if (conn->fi->state != ST_CLEARING)
 			osmo_fsm_inst_state_chg(fi, ST_CLEARING, 60, -4);
 		LOGPFSML(fi, LOGL_DEBUG, "Releasing all lchans (if any) after BSSMAP Clear Command\n");
-		gscon_release_lchans(conn, true, bsc_gsm48_rr_cause_from_gsm0808_cause(ccd->cause_0808));
+		gscon_release_lchans(conn, true, bsc_gsm48_rr_cause_from_gsm0808_cause(*cause_0808));
 		/* FIXME: Release all terestrial resources in ST_CLEARING */
 		/* According to 3GPP 48.008 3.1.9.1. "The BSS need not wait for the radio channel
 		 * release to be completed or for the guard timer to expire before returning the
@@ -841,7 +868,7 @@ static void gscon_fsm_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *d
 		/* Close MGCP connections */
 		osmo_mgcpc_ep_clear(conn->user_plane.mgw_endpoint);
 
-		rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_CLEAR_COMPLETE]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_CLEAR_COMPLETE));
 		gscon_sigtran_send(conn, gsm0808_create_clear_complete());
 		break;
 	case GSCON_EV_A_DISC_IND:
@@ -868,14 +895,26 @@ static void gscon_fsm_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *d
 		break;
 	case GSCON_EV_A_COMMON_ID_IND:
 		OSMO_ASSERT(data);
-		mi_imsi = data;
+		tp = data;
+		if (osmo_mobile_identity_decode(&mi_imsi, TLVP_VAL(tp, GSM0808_IE_IMSI), TLVP_LEN(tp, GSM0808_IE_IMSI), false)
+		    || mi_imsi.type != GSM_MI_TYPE_IMSI) {
+			LOGPFSML(fi, LOGL_ERROR, "CommonID: could not parse IMSI\n");
+			return;
+		}
 		if (!conn->bsub)
-			conn->bsub = bsc_subscr_find_or_create_by_imsi(conn->network->bsc_subscribers, mi_imsi->imsi,
+			conn->bsub = bsc_subscr_find_or_create_by_imsi(conn->network->bsc_subscribers, mi_imsi.imsi,
 								       BSUB_USE_CONN);
 		else {
 			/* we already have a bsc_subscr associated; maybe that subscriber has no IMSI yet? */
 			if (!conn->bsub->imsi[0])
-				bsc_subscr_set_imsi(conn->bsub, mi_imsi->imsi);
+				bsc_subscr_set_imsi(conn->bsub, mi_imsi.imsi);
+		}
+		if (TLVP_PRESENT(tp, GSM0808_IE_LAST_USED_EUTRAN_PLMN_ID)) {
+			conn->fast_return.allowed = true; /* Always allowed for CSFB */
+			conn->fast_return.last_eutran_plmn_valid = true;
+			osmo_plmn_from_bcd(TLVP_VAL(tp, GSM0808_IE_LAST_USED_EUTRAN_PLMN_ID), &conn->fast_return.last_eutran_plmn);
+			LOGPFSML(fi, LOGL_DEBUG, "subscr comes from E-UTRAN PLMN %s\n",
+				 osmo_plmn_name(&conn->fast_return.last_eutran_plmn));
 		}
 		gscon_update_id(conn);
 		break;
@@ -916,8 +955,14 @@ static void gscon_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cau
 static void gscon_pre_term(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
 {
 	struct gsm_subscriber_connection *conn = fi->priv;
+	struct mgcp_client *mgcp_client;
+
+	/* Put MGCP client back into MGW pool */
+	mgcp_client = osmo_mgcpc_ep_client(conn->user_plane.mgw_endpoint);
+	mgcp_client_pool_put(mgcp_client);
 
 	osmo_mgcpc_ep_clear(conn->user_plane.mgw_endpoint);
+	conn->user_plane.mgw_endpoint = NULL;
 
 	if (conn->lcls.fi) {
 		/* request termination of LCLS FSM */
@@ -933,7 +978,7 @@ static void gscon_pre_term(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause ca
 	/* drop pending messages */
 	gscon_dtap_queue_flush(conn, 0);
 
-	penalty_timers_free(&conn->hodec2.penalty_timers);
+	penalty_timers_clear(&conn->hodec2.penalty_timers, NULL);
 }
 
 static int gscon_timer_cb(struct osmo_fsm_inst *fi)
@@ -1004,7 +1049,7 @@ struct gsm_subscriber_connection *bsc_subscr_con_allocate(struct gsm_network *ne
 
 	conn->network = net;
 	INIT_LLIST_HEAD(&conn->dtap_queue);
-	/* BTW, penalty timers will be initialized on-demand. */
+	INIT_LLIST_HEAD(&conn->hodec2.penalty_timers);
 	conn->sccp.conn_id = -1;
 
 	/* don't allocate from 'conn' context, as gscon_cleanup() will call talloc_free(conn) before
@@ -1123,7 +1168,6 @@ static void gsm0808_send_rsl_dtap(struct gsm_subscriber_connection *conn,
 
 	sapi = link_id & 0x7;
 	msg->lchan = conn->lchan;
-	msg->dst = msg->lchan->ts->trx->rsl_link;
 
 	/* If we are on a TCH and need to submit a SMS (on SAPI=3) we need to use the SACH */
 	if (allow_sacch && sapi != 0) {
