@@ -1,5 +1,5 @@
 /* (C) 2008-2018 by Harald Welte <laforge@gnumonks.org>
- * (C) 2020 by sysmocom s.f.m.c. GmbH <info@sysmocom.de>
+ * (C) 2021 by sysmocom s.f.m.c. GmbH <info@sysmocom.de>
  *
  * All Rights Reserved
  *
@@ -67,7 +67,7 @@ const struct value_string bts_type_names[_NUM_GSM_BTS_TYPE + 1] = {
 	{ GSM_BTS_TYPE_NANOBTS,		"nanobts" },
 	{ GSM_BTS_TYPE_RBS2000,		"rbs2000" },
 	{ GSM_BTS_TYPE_NOKIA_SITE,	"nokia_site" },
-	{ GSM_BTS_TYPE_OSMOBTS,		"sysmobts" },
+	{ GSM_BTS_TYPE_OSMOBTS,		"osmo-bts" },
 	{ 0, NULL }
 };
 
@@ -77,7 +77,7 @@ const struct value_string bts_type_descs[_NUM_GSM_BTS_TYPE+1] = {
 	{ GSM_BTS_TYPE_NANOBTS,		"ip.access nanoBTS or compatible" },
 	{ GSM_BTS_TYPE_RBS2000,		"Ericsson RBS2000 Series" },
 	{ GSM_BTS_TYPE_NOKIA_SITE,	"Nokia {Metro,Ultra,In}Site" },
-	{ GSM_BTS_TYPE_OSMOBTS,		"sysmocom sysmoBTS" },
+	{ GSM_BTS_TYPE_OSMOBTS,		"Osmocom Base Transceiver Station" },
 	{ 0,				NULL }
 };
 
@@ -211,6 +211,51 @@ struct gsm_bts *gsm_bts_alloc(struct gsm_network *net, struct gsm_bts_sm *bts_sm
 	}
 	bts->bts_statg = osmo_stat_item_group_alloc(bts, &bts_statg_desc, bts->nr);
 
+	bts->all_allocated_sdcch = (struct osmo_time_cc){
+		.cfg = {
+			.gran_usec = 1*1000000,
+			.forget_sum_usec = 60*1000000,
+			.rate_ctr = rate_ctr_group_get_ctr(bts->bts_ctrs, BTS_CTR_ALL_ALLOCATED_SDCCH),
+			.T_gran = -16,
+			.T_round_threshold = -17,
+			.T_forget_sum = -18,
+			.T_defs = net->T_defs,
+		},
+	};
+	bts->all_allocated_static_sdcch = (struct osmo_time_cc){
+		.cfg = {
+			.gran_usec = 1*1000000,
+			.forget_sum_usec = 60*1000000,
+			.rate_ctr = rate_ctr_group_get_ctr(bts->bts_ctrs, BTS_CTR_ALL_ALLOCATED_STATIC_SDCCH),
+			.T_gran = -16,
+			.T_round_threshold = -17,
+			.T_forget_sum = -18,
+			.T_defs = net->T_defs,
+		},
+	};
+	bts->all_allocated_tch = (struct osmo_time_cc){
+		.cfg = {
+			.gran_usec = 1*1000000,
+			.forget_sum_usec = 60*1000000,
+			.rate_ctr = rate_ctr_group_get_ctr(bts->bts_ctrs, BTS_CTR_ALL_ALLOCATED_TCH),
+			.T_gran = -16,
+			.T_round_threshold = -17,
+			.T_forget_sum = -18,
+			.T_defs = net->T_defs,
+		},
+	};
+	bts->all_allocated_static_tch = (struct osmo_time_cc){
+		.cfg = {
+			.gran_usec = 1*1000000,
+			.forget_sum_usec = 60*1000000,
+			.rate_ctr = rate_ctr_group_get_ctr(bts->bts_ctrs, BTS_CTR_ALL_ALLOCATED_STATIC_TCH),
+			.T_gran = -16,
+			.T_round_threshold = -17,
+			.T_forget_sum = -18,
+			.T_defs = net->T_defs,
+		},
+	};
+
 	/* create our primary TRX */
 	bts->c0 = gsm_bts_trx_alloc(bts);
 	if (!bts->c0) {
@@ -245,6 +290,7 @@ struct gsm_bts *gsm_bts_alloc(struct gsm_network *net, struct gsm_bts_sm *bts_sm
 	bts->neigh_list_manual_mode = NL_MODE_AUTOMATIC;
 	bts->early_classmark_allowed_3g = true; /* 3g Early Classmark Sending controlled by bts->early_classmark_allowed param */
 	bts->si_unused_send_empty = true;
+	bts->chan_alloc_allow_tch_for_signalling = true;
 	bts->si_common.cell_sel_par.cell_resel_hyst = 2; /* 4 dB */
 	bts->si_common.cell_sel_par.rxlev_acc_min = 0;
 	bts->si_common.si2quater_neigh_list.arfcn = bts->si_common.data.earfcn_list;
@@ -265,15 +311,19 @@ struct gsm_bts *gsm_bts_alloc(struct gsm_network *net, struct gsm_bts_sm *bts_sm
 	bts->si_common.rach_control.tx_integer = 9;  /* 12 slots spread - 217/115 slots delay */
 	bts->si_common.rach_control.max_trans = 3; /* 7 retransmissions */
 	bts->si_common.rach_control.t2 = 4; /* no emergency calls */
+	bts->si_common.chan_desc.mscr = 1; /* Indicate R99 MSC in SI3 */
 	bts->si_common.chan_desc.att = 1; /* attachment required */
 	bts->si_common.chan_desc.bs_pa_mfrms = RSL_BS_PA_MFRMS_5; /* paging frames */
 	bts->si_common.chan_desc.bs_ag_blks_res = 1; /* reserved AGCH blocks */
 	bts->si_common.chan_desc.t3212 = osmo_tdef_get(net->T_defs, 3212, OSMO_TDEF_CUSTOM, -1);
+	bts->si_common.cell_options.pwrc = 0; /* PWRC not set */
+	bts->si_common.cell_sel_par.acs = 0;
+	bts->si_common.ncc_permitted = 0xff;
 	gsm_bts_set_radio_link_timeout(bts, 32); /* Use RADIO LINK TIMEOUT of 32 */
 
 	INIT_LLIST_HEAD(&bts->abis_queue);
 	INIT_LLIST_HEAD(&bts->loc_list);
-	INIT_LLIST_HEAD(&bts->local_neighbors);
+	INIT_LLIST_HEAD(&bts->neighbors);
 	INIT_LLIST_HEAD(&bts->oml_fail_rep);
 	INIT_LLIST_HEAD(&bts->chan_rqd_queue);
 
@@ -347,15 +397,26 @@ struct gsm_bts *gsm_bts_alloc(struct gsm_network *net, struct gsm_bts_sm *bts_sm
 	acc_mgr_init(&bts->acc_mgr, bts);
 	acc_ramp_init(&bts->acc_ramp, bts);
 
-	bts->repeated_acch_policy.rxqual = 4;
+	/* Default RxQual threshold for ACCH repetition/overpower */
+	bts->rep_acch_cap.rxqual = 4;
+	bts->top_acch_cap.rxqual = 4;
+
+	/* Permit ACCH overpower only for speech channels using AMR */
+	bts->top_acch_chan_mode = TOP_ACCH_CHAN_MODE_SPEECH_V3;
 
 	/* MS Power Control parameters (defaults) */
-	bts->ms_power_ctrl = power_ctrl_params_def;
-	bts->ms_power_ctrl.dir = GSM_PWR_CTRL_DIR_UL;
+	power_ctrl_params_def_reset(&bts->ms_power_ctrl, GSM_PWR_CTRL_DIR_UL);
 
 	/* BS Power Control parameters (defaults) */
-	bts->bs_power_ctrl = power_ctrl_params_def;
-	bts->bs_power_ctrl.dir = GSM_PWR_CTRL_DIR_DL;
+	power_ctrl_params_def_reset(&bts->bs_power_ctrl, GSM_PWR_CTRL_DIR_DL);
+
+	/* Interference Measurement Parameters (defaults) */
+	bts->interf_meas_params_cfg = interf_meas_params_def;
+
+	bts->rach_max_delay = 63;
+
+	/* SRVCC is enabled by default */
+	bts->srvcc_fast_return_allowed = true;
 
 	return bts;
 }
@@ -414,57 +475,32 @@ bool gsm_bts_matches_cell_id(const struct gsm_bts *bts, const struct gsm0808_cel
 	}
 }
 
-static struct gsm_bts_ref *gsm_bts_ref_find(const struct llist_head *list, const struct gsm_bts *bts)
+/* Return a LAC+CI cell identity for the given BTS.
+ * (For matching a BTS within the local BSS, the PLMN code is not important.) */
+void gsm_bts_cell_id(struct gsm0808_cell_id *cell_id, const struct gsm_bts *bts)
 {
-	struct gsm_bts_ref *ref;
-	if (!bts)
-		return NULL;
-	llist_for_each_entry(ref, list, entry) {
-		if (ref->bts == bts)
-			return ref;
-	}
-	return NULL;
+	*cell_id = (struct gsm0808_cell_id){
+		.id_discr = CELL_IDENT_LAC_AND_CI,
+		.id.lac_and_ci = {
+			.lac = bts->location_area_code,
+			.ci = bts->cell_identity,
+		},
+	};
 }
 
-/* Add a BTS reference to the local_neighbors list.
- * Return 1 if added, 0 if such an entry already existed, and negative on errors. */
-int gsm_bts_local_neighbor_add(struct gsm_bts *bts, struct gsm_bts *neighbor)
+/* Same as gsm_bts_cell_id(), but return in a single-entry gsm0808_cell_id_list2. Useful for e.g.
+ * gsm0808_cell_id_list_add() and gsm0808_cell_id_lists_same(). */
+void gsm_bts_cell_id_list(struct gsm0808_cell_id_list2 *cell_id_list, const struct gsm_bts *bts)
 {
-	struct gsm_bts_ref *ref;
-	if (!bts || !neighbor)
-		return -ENOMEM;
-
-	if (bts == neighbor)
-		return -EINVAL;
-
-	/* Already got this entry? */
-	ref = gsm_bts_ref_find(&bts->local_neighbors, neighbor);
-	if (ref)
-		return 0;
-
-	ref = talloc_zero(bts, struct gsm_bts_ref);
-	if (!ref)
-		return -ENOMEM;
-	ref->bts = neighbor;
-	llist_add_tail(&ref->entry, &bts->local_neighbors);
-	return 1;
-}
-
-/* Remove a BTS reference from the local_neighbors list.
- * Return 1 if removed, 0 if no such entry existed, and negative on errors. */
-int gsm_bts_local_neighbor_del(struct gsm_bts *bts, const struct gsm_bts *neighbor)
-{
-	struct gsm_bts_ref *ref;
-	if (!bts || !neighbor)
-		return -ENOMEM;
-
-	ref = gsm_bts_ref_find(&bts->local_neighbors, neighbor);
-	if (!ref)
-		return 0;
-
-	llist_del(&ref->entry);
-	talloc_free(ref);
-	return 1;
+	struct gsm0808_cell_id cell_id;
+	struct gsm0808_cell_id_list2 add;
+	int rc;
+	gsm_bts_cell_id(&cell_id, bts);
+	gsm0808_cell_id_to_list(&add, &cell_id);
+	/* Since the target list is empty, this should always succeed. */
+	(*cell_id_list) = (struct gsm0808_cell_id_list2){};
+	rc = gsm0808_cell_id_list_add(cell_id_list, &add);
+	OSMO_ASSERT(rc > 0);
 }
 
 /* return the gsm_lchan for the CBCH (if it exists at all) */
@@ -579,16 +615,19 @@ struct gsm_bts_trx *gsm_bts_trx_num(const struct gsm_bts *bts, int num)
 	return NULL;
 }
 
+void bts_store_uptime(struct gsm_bts *bts)
+{
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_UPTIME_SECONDS), bts_uptime(bts));
+}
+
 unsigned long long bts_uptime(const struct gsm_bts *bts)
 {
 	struct timespec tp;
 
-	if (!bts->uptime || !bts->oml_link) {
-		LOGP(DNM, LOGL_ERROR, "BTS %u OML link uptime unavailable\n", bts->nr);
+	if (!bts->uptime || !bts->oml_link)
 		return 0;
-	}
 
-	if (clock_gettime(CLOCK_MONOTONIC, &tp) != 0) {
+	if (osmo_clock_gettime(CLOCK_MONOTONIC, &tp) != 0) {
 		LOGP(DNM, LOGL_ERROR, "BTS %u uptime computation failure: %s\n", bts->nr, strerror(errno));
 		return 0;
 	}
@@ -728,19 +767,6 @@ void gsm_bts_all_ts_dispatch(struct gsm_bts *bts, uint32_t ts_ev, void *data)
 		gsm_trx_all_ts_dispatch(trx, ts_ev, data);
 }
 
-
-/* Count number of free TS of given pchan type */
-int bts_count_free_ts(struct gsm_bts *bts, enum gsm_phys_chan_config pchan)
-{
-	struct gsm_bts_trx *trx;
-	int count = 0;
-
-	llist_for_each_entry(trx, &bts->trx_list, list)
-		count += trx_count_free_ts(trx, pchan);
-
-	return count;
-}
-
 /* set all system information types for a BTS */
 int gsm_bts_set_system_infos(struct gsm_bts *bts)
 {
@@ -761,6 +787,75 @@ int gsm_bts_set_system_infos(struct gsm_bts *bts)
 	return 0;
 }
 
+int gsm_bts_set_c0_power_red(struct gsm_bts *bts, const uint8_t red)
+{
+	struct gsm_bts_trx *c0 = bts->c0;
+	unsigned int tn;
+	int rc;
+
+	if (!osmo_bts_has_feature(&bts->features, BTS_FEAT_BCCH_POWER_RED))
+		return -ENOTSUP;
+	if (bts->model->power_ctrl_set_c0_power_red == NULL)
+		return -ENOTSUP;
+
+	rc = bts->model->power_ctrl_set_c0_power_red(bts, red);
+	if (rc != 0)
+		return rc;
+
+	/* Timeslot 0 is always transmitting BCCH/CCCH */
+	c0->ts[0].c0_max_power_red_db = 0;
+
+	for (tn = 1; tn < ARRAY_SIZE(c0->ts); tn++) {
+		struct gsm_bts_trx_ts *ts = &c0->ts[tn];
+		struct gsm_bts_trx_ts *prev = ts - 1;
+
+		switch (ts->pchan_is) {
+		/* Not allowed on CCCH/BCCH */
+		case GSM_PCHAN_CCCH:
+			/* Preceeding timeslot shall not exceed 2 dB */
+			if (prev->c0_max_power_red_db > 0)
+				prev->c0_max_power_red_db = 2;
+			/* fall-through */
+		/* Not recommended on SDCCH/8 */
+		case GSM_PCHAN_SDCCH8_SACCH8C:
+		case GSM_PCHAN_SDCCH8_SACCH8C_CBCH:
+			ts->c0_max_power_red_db = 0;
+			break;
+		default:
+			ts->c0_max_power_red_db = red;
+			break;
+		}
+	}
+
+	/* Timeslot 7 is always preceding BCCH/CCCH */
+	if (c0->ts[7].c0_max_power_red_db > 0)
+		c0->ts[7].c0_max_power_red_db = 2;
+
+	bts->c0_max_power_red_db = red;
+
+	return 0;
+}
+
+void gsm_bts_stats_reset(struct gsm_bts *bts)
+{
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_CCCH_SDCCH4_USED), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_CCCH_SDCCH4_TOTAL), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_TCH_F_USED), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_TCH_F_TOTAL), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_TCH_H_USED), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_TCH_H_TOTAL), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_SDCCH8_USED), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_SDCCH8_TOTAL), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_TCH_F_PDCH_USED), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_TCH_F_PDCH_TOTAL), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_CCCH_SDCCH4_CBCH_USED), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_CCCH_SDCCH4_CBCH_TOTAL), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_SDCCH8_CBCH_USED), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_SDCCH8_CBCH_TOTAL), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_OSMO_DYN_USED), 0);
+	osmo_stat_item_set(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_CHAN_OSMO_DYN_TOTAL), 0);
+}
+
 const struct rate_ctr_desc bts_ctr_description[] = {
 	[BTS_CTR_CHREQ_TOTAL] = \
 		{ "chreq:total",
@@ -789,9 +884,33 @@ const struct rate_ctr_desc bts_ctr_description[] = {
 	[BTS_CTR_CHREQ_SUCCESSFUL] = \
 		{ "chreq:successful",
 		  "Successful channel requests (immediate assign sent)" },
+	[BTS_CTR_CHREQ_SUCCESSFUL_EMERG] = \
+		{ "chreq:successful_emerg",
+		  "Sent Immediate Assignment for EMERG" },
+	[BTS_CTR_CHREQ_SUCCESSFUL_CALL] = \
+		{ "chreq:successful_call",
+		  "Sent Immediate Assignment for CALL" },
+	[BTS_CTR_CHREQ_SUCCESSFUL_LOCATION_UPD] = \
+		{ "chreq:successful_location_upd",
+		  "Sent Immediate Assignment for LOCATION_UPD" },
+	[BTS_CTR_CHREQ_SUCCESSFUL_PAG] = \
+		{ "chreq:successful_pag",
+		  "Sent Immediate Assignment for PAG" },
+	[BTS_CTR_CHREQ_SUCCESSFUL_PDCH] = \
+		{ "chreq:successful_pdch",
+		  "Sent Immediate Assignment for PDCH" },
+	[BTS_CTR_CHREQ_SUCCESSFUL_OTHER] = \
+		{ "chreq:successful_other",
+		  "Sent Immediate Assignment for OTHER" },
+	[BTS_CTR_CHREQ_SUCCESSFUL_UNKNOWN] = \
+		{ "chreq:successful_unknown",
+		  "Sent Immediate Assignment for UNKNOWN" },
 	[BTS_CTR_CHREQ_NO_CHANNEL] = \
 		{ "chreq:no_channel",
 		  "Sent to MS no channel available" },
+	[BTS_CTR_CHREQ_MAX_DELAY_EXCEEDED] = \
+		{ "chreq:max_delay_exceeded",
+		  "Received channel requests with greater than permitted access delay" },
 	[BTS_CTR_CHAN_RF_FAIL] = \
 		{ "chan:rf_fail",
 		  "Received a RF failure indication from BTS" },
@@ -905,6 +1024,9 @@ const struct rate_ctr_desc bts_ctr_description[] = {
 	[BTS_CTR_LCHAN_BORKEN_EV_TEARDOWN] = \
 		{ "lchan_borken:event:teardown",
 		  "lchan in a BORKEN state is shutting down (BTS disconnected?)" },
+	[BTS_CTR_LCHAN_BORKEN_EV_TS_ERROR] = \
+		{ "lchan_borken:event:ts_error",
+		  "LCHAN_EV_TS_ERROR received in a BORKEN state" },
 	[BTS_CTR_TS_BORKEN_FROM_NOT_INITIALIZED] = \
 		{ "ts_borken:from_state:not_initialized",
 		  "Transitions from TS NOT_INITIALIZED state to BORKEN state" },
@@ -1059,10 +1181,10 @@ const struct rate_ctr_desc bts_ctr_description[] = {
 
 	[BTS_CTR_INTRA_BSC_HO_ATTEMPTED] = \
 		{ "intra_bsc_ho:attempted",
-		  "Intra-BSC handover attempts" },
+		  "Intra-BSC inter-cell handover attempts" },
 	[BTS_CTR_INTRA_BSC_HO_COMPLETED] = \
 		{ "intra_bsc_ho:completed",
-		  "Intra-BSC handover completed" },
+		  "Intra-BSC inter-cell handover completed" },
 	[BTS_CTR_INTRA_BSC_HO_STOPPED] = \
 		{ "intra_bsc_ho:stopped",
 		  "Connection ended during HO" },
@@ -1077,7 +1199,29 @@ const struct rate_ctr_desc bts_ctr_description[] = {
 		  "Received Handover Fail messages" },
 	[BTS_CTR_INTRA_BSC_HO_ERROR] = \
 		{ "intra_bsc_ho:error",
-		  "Re-assignment failed for other reason" },
+		  "Intra-BSC inter-cell HO failed for other reason" },
+
+	[BTS_CTR_INCOMING_INTRA_BSC_HO_ATTEMPTED] = \
+		{ "incoming_intra_bsc_ho:attempted",
+		  "Incoming intra-BSC inter-cell handover attempts" },
+	[BTS_CTR_INCOMING_INTRA_BSC_HO_COMPLETED] = \
+		{ "incoming_intra_bsc_ho:completed",
+		  "Incoming intra-BSC inter-cell handover completed" },
+	[BTS_CTR_INCOMING_INTRA_BSC_HO_STOPPED] = \
+		{ "incoming_intra_bsc_ho:stopped",
+		  "Connection ended during HO" },
+	[BTS_CTR_INCOMING_INTRA_BSC_HO_NO_CHANNEL] = \
+		{ "incoming_intra_bsc_ho:no_channel",
+		  "Failure to allocate lchan for HO" },
+	[BTS_CTR_INCOMING_INTRA_BSC_HO_TIMEOUT] = \
+		{ "incoming_intra_bsc_ho:timeout",
+		  "Handover timed out" },
+	[BTS_CTR_INCOMING_INTRA_BSC_HO_FAILED] = \
+		{ "incoming_intra_bsc_ho:failed",
+		  "Received Handover Fail messages" },
+	[BTS_CTR_INCOMING_INTRA_BSC_HO_ERROR] = \
+		{ "incoming_intra_bsc_ho:error",
+		  "Incoming intra-BSC inter-cell HO failed for other reason" },
 
 	[BTS_CTR_INTER_BSC_HO_OUT_ATTEMPTED] = \
 		{ "interbsc_ho_out:attempted",
@@ -1118,6 +1262,115 @@ const struct rate_ctr_desc bts_ctr_description[] = {
 	[BTS_CTR_INTER_BSC_HO_IN_ERROR] = \
 		{ "interbsc_ho_in:error",
 		  "Handover from remote BSS failed for other reason" },
+
+	[BTS_CTR_SRVCC_ATTEMPTED] = \
+		{ "srvcc:attempted",
+		  "Intra-BSC handover attempts" },
+	[BTS_CTR_SRVCC_COMPLETED] = \
+		{ "srvcc:completed",
+		  "Intra-BSC handover completed" },
+	[BTS_CTR_SRVCC_STOPPED] = \
+		{ "srvcc:stopped",
+		  "Connection ended during HO" },
+	[BTS_CTR_SRVCC_NO_CHANNEL] = \
+		{ "srvcc:no_channel",
+		  "Failure to allocate lchan for HO" },
+	[BTS_CTR_SRVCC_TIMEOUT] = \
+		{ "srvcc:timeout",
+		   "Handover timed out" },
+	[BTS_CTR_SRVCC_FAILED] = \
+		{ "srvcc:failed",
+		  "Received Handover Fail messages" },
+	[BTS_CTR_SRVCC_ERROR] = \
+		{ "srvcc:error",
+		  "Re-assignment failed for other reason" },
+	[BTS_CTR_ALL_ALLOCATED_SDCCH] = \
+		{ "all_allocated:sdcch",
+		  "Cumulative counter of seconds where all SDCCH channels were allocated" },
+	[BTS_CTR_ALL_ALLOCATED_STATIC_SDCCH] = \
+		{ "all_allocated:static_sdcch",
+		  "Cumulative counter of seconds where all non-dynamic SDCCH channels were allocated" },
+	[BTS_CTR_ALL_ALLOCATED_TCH] = \
+		{ "all_allocated:tch",
+		  "Cumulative counter of seconds where all TCH channels were allocated" },
+	[BTS_CTR_ALL_ALLOCATED_STATIC_TCH] = \
+		{ "all_allocated:static_tch",
+		  "Cumulative counter of seconds where all non-dynamic TCH channels were allocated" },
+
+	[BTS_CTR_CM_SERV_REJ] = \
+		{ "cm_serv_rej", "MSC sent CM Service Reject" },
+	[BTS_CTR_CM_SERV_REJ_IMSI_UNKNOWN_IN_HLR] = \
+		{ "cm_serv_rej:imsi_unknown_in_hlr",
+		  "MSC sent CM Service Reject with cause IMSI_UNKNOWN_IN_HLR" },
+	[BTS_CTR_CM_SERV_REJ_ILLEGAL_MS] = \
+		{ "cm_serv_rej:illegal_ms",
+		  "MSC sent CM Service Reject with cause ILLEGAL_MS" },
+	[BTS_CTR_CM_SERV_REJ_IMSI_UNKNOWN_IN_VLR] = \
+		{ "cm_serv_rej:imsi_unknown_in_vlr",
+		  "MSC sent CM Service Reject with cause IMSI_UNKNOWN_IN_VLR" },
+	[BTS_CTR_CM_SERV_REJ_IMEI_NOT_ACCEPTED] = \
+		{ "cm_serv_rej:imei_not_accepted",
+		  "MSC sent CM Service Reject with cause IMEI_NOT_ACCEPTED" },
+	[BTS_CTR_CM_SERV_REJ_ILLEGAL_ME] = \
+		{ "cm_serv_rej:illegal_me",
+		  "MSC sent CM Service Reject with cause ILLEGAL_ME" },
+	[BTS_CTR_CM_SERV_REJ_PLMN_NOT_ALLOWED] = \
+		{ "cm_serv_rej:plmn_not_allowed",
+		  "MSC sent CM Service Reject with cause PLMN_NOT_ALLOWED" },
+	[BTS_CTR_CM_SERV_REJ_LOC_NOT_ALLOWED] = \
+		{ "cm_serv_rej:loc_not_allowed",
+		  "MSC sent CM Service Reject with cause LOC_NOT_ALLOWED" },
+	[BTS_CTR_CM_SERV_REJ_ROAMING_NOT_ALLOWED] = \
+		{ "cm_serv_rej:roaming_not_allowed",
+		  "MSC sent CM Service Reject with cause ROAMING_NOT_ALLOWED" },
+	[BTS_CTR_CM_SERV_REJ_NETWORK_FAILURE] = \
+		{ "cm_serv_rej:network_failure",
+		  "MSC sent CM Service Reject with cause NETWORK_FAILURE" },
+	[BTS_CTR_CM_SERV_REJ_SYNCH_FAILURE] = \
+		{ "cm_serv_rej:synch_failure",
+		  "MSC sent CM Service Reject with cause SYNCH_FAILURE" },
+	[BTS_CTR_CM_SERV_REJ_CONGESTION] = \
+		{ "cm_serv_rej:congestion",
+		  "MSC sent CM Service Reject with cause CONGESTION" },
+	[BTS_CTR_CM_SERV_REJ_SRV_OPT_NOT_SUPPORTED] = \
+		{ "cm_serv_rej:srv_opt_not_supported",
+		  "MSC sent CM Service Reject with cause SRV_OPT_NOT_SUPPORTED" },
+	[BTS_CTR_CM_SERV_REJ_RQD_SRV_OPT_NOT_SUPPORTED] = \
+		{ "cm_serv_rej:rqd_srv_opt_not_supported",
+		  "MSC sent CM Service Reject with cause RQD_SRV_OPT_NOT_SUPPORTED" },
+	[BTS_CTR_CM_SERV_REJ_SRV_OPT_TMP_OUT_OF_ORDER] = \
+		{ "cm_serv_rej:srv_opt_tmp_out_of_order",
+		  "MSC sent CM Service Reject with cause SRV_OPT_TMP_OUT_OF_ORDER" },
+	[BTS_CTR_CM_SERV_REJ_CALL_CAN_NOT_BE_IDENTIFIED] = \
+		{ "cm_serv_rej:call_can_not_be_identified",
+		  "MSC sent CM Service Reject with cause CALL_CAN_NOT_BE_IDENTIFIED" },
+	[BTS_CTR_CM_SERV_REJ_INCORRECT_MESSAGE] = \
+		{ "cm_serv_rej:incorrect_message",
+		  "MSC sent CM Service Reject with cause INCORRECT_MESSAGE" },
+	[BTS_CTR_CM_SERV_REJ_INVALID_MANDANTORY_INF] = \
+		{ "cm_serv_rej:invalid_mandantory_inf",
+		  "MSC sent CM Service Reject with cause INVALID_MANDANTORY_INF" },
+	[BTS_CTR_CM_SERV_REJ_MSG_TYPE_NOT_IMPLEMENTED] = \
+		{ "cm_serv_rej:msg_type_not_implemented",
+		  "MSC sent CM Service Reject with cause MSG_TYPE_NOT_IMPLEMENTED" },
+	[BTS_CTR_CM_SERV_REJ_MSG_TYPE_NOT_COMPATIBLE] = \
+		{ "cm_serv_rej:msg_type_not_compatible",
+		  "MSC sent CM Service Reject with cause MSG_TYPE_NOT_COMPATIBLE" },
+	[BTS_CTR_CM_SERV_REJ_INF_ELEME_NOT_IMPLEMENTED] = \
+		{ "cm_serv_rej:inf_eleme_not_implemented",
+		  "MSC sent CM Service Reject with cause INF_ELEME_NOT_IMPLEMENTED" },
+	[BTS_CTR_CM_SERV_REJ_CONDTIONAL_IE_ERROR] = \
+		{ "cm_serv_rej:condtional_ie_error",
+		  "MSC sent CM Service Reject with cause CONDTIONAL_IE_ERROR" },
+	[BTS_CTR_CM_SERV_REJ_MSG_NOT_COMPATIBLE] = \
+		{ "cm_serv_rej:msg_not_compatible",
+		  "MSC sent CM Service Reject with cause MSG_NOT_COMPATIBLE" },
+	[BTS_CTR_CM_SERV_REJ_PROTOCOL_ERROR] = \
+		{ "cm_serv_rej:protocol_error",
+		  "MSC sent CM Service Reject with cause PROTOCOL_ERROR" },
+	[BTS_CTR_CM_SERV_REJ_RETRY_IN_NEW_CELL] = \
+		{ "cm_serv_rej:retry_in_new_cell",
+		  "MSC sent CM Service Reject with cause 00110000..00111111, Retry upon entry in a new cell" },
 };
 
 const struct rate_ctr_group_desc bts_ctrg_desc = {
@@ -1129,93 +1382,97 @@ const struct rate_ctr_group_desc bts_ctrg_desc = {
 };
 
 const struct osmo_stat_item_desc bts_stat_desc[] = {
+	[BTS_STAT_UPTIME_SECONDS] = \
+		{ "uptime:seconds",
+		  "Seconds of uptime",
+		  "s", 60, 0 },
 	[BTS_STAT_CHAN_LOAD_AVERAGE] = \
 		{ "chanloadavg",
 		  "Channel load average",
-		  "%", 16, 0 },
+		  "%", 60, 0 },
 	[BTS_STAT_CHAN_CCCH_SDCCH4_USED] = \
 		{ "chan_ccch_sdcch4:used",
 		  "Number of CCCH+SDCCH4 channels used",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_CCCH_SDCCH4_TOTAL] = \
 		{ "chan_ccch_sdcch4:total",
 		  "Number of CCCH+SDCCH4 channels total",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_TCH_F_USED] = \
 		{ "chan_tch_f:used",
 		  "Number of TCH/F channels used",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_TCH_F_TOTAL] = \
 		{ "chan_tch_f:total",
 		  "Number of TCH/F channels total",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_TCH_H_USED] = \
 		{ "chan_tch_h:used",
 		  "Number of TCH/H channels used",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_TCH_H_TOTAL] = \
 		{ "chan_tch_h:total",
 		  "Number of TCH/H channels total",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_SDCCH8_USED] = \
 		{ "chan_sdcch8:used",
 		  "Number of SDCCH8 channels used",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_SDCCH8_TOTAL] = \
 		{ "chan_sdcch8:total",
 		  "Number of SDCCH8 channels total",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_TCH_F_PDCH_USED] = \
 		{ "chan_tch_f_pdch:used",
 		  "Number of TCH/F_PDCH channels used",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_TCH_F_PDCH_TOTAL] = \
 		{ "chan_tch_f_pdch:total",
 		  "Number of TCH/F_PDCH channels total",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_CCCH_SDCCH4_CBCH_USED] = \
 		{ "chan_ccch_sdcch4_cbch:used",
 		  "Number of CCCH+SDCCH4+CBCH channels used",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_CCCH_SDCCH4_CBCH_TOTAL] = \
 		{ "chan_ccch_sdcch4_cbch:total",
 		  "Number of CCCH+SDCCH4+CBCH channels total",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_SDCCH8_CBCH_USED] = \
 		{ "chan_sdcch8_cbch:used",
 		  "Number of SDCCH8+CBCH channels used",
-		  "", 16, 0 },
+		  "", 60, 0 },
 	[BTS_STAT_CHAN_SDCCH8_CBCH_TOTAL] = \
 		{ "chan_sdcch8_cbch:total",
 		  "Number of SDCCH8+CBCH channels total",
-		  "", 16, 0 },
-	[BTS_STAT_CHAN_TCH_F_TCH_H_PDCH_USED] = \
-		{ "chan_tch_f_tch_h_pdch:used",
-		  "Number of TCH/F_TCH/H_PDCH channels used",
-		  "", 16, 0 },
-	[BTS_STAT_CHAN_TCH_F_TCH_H_PDCH_TOTAL] = \
-		{ "chan_tch_f_tch_h_pdch:total",
-		  "Number of TCH/F_TCH/H_PDCH channels total",
-		  "", 16, 0 },
+		  "", 60, 0 },
+	[BTS_STAT_CHAN_OSMO_DYN_USED] = \
+		{ "chan_osmo_dyn:used",
+		  "Number of TCH/F_TCH/H_SDCCH8_PDCH channels used",
+		  "", 60, 0 },
+	[BTS_STAT_CHAN_OSMO_DYN_TOTAL] = \
+		{ "chan_osmo_dyn:total",
+		  "Number of TCH/F_TCH/H_SDCCH8_PDCH channels total",
+		  "", 60, 0 },
 	[BTS_STAT_T3122] = \
 		{ "T3122",
 		  "T3122 IMMEDIATE ASSIGNMENT REJECT wait indicator",
-		  "s", 16, GSM_T3122_DEFAULT },
+		  "s", 60, GSM_T3122_DEFAULT },
 	[BTS_STAT_RACH_BUSY] = \
 		{ "rach_busy",
 		  "RACH slots with signal above threshold",
-		  "%", 16, 0 },
+		  "%", 60, 0 },
 	[BTS_STAT_RACH_ACCESS] = \
 		{ "rach_access",
 		  "RACH slots with access bursts in them",
-		  "%", 16, 0 },
+		  "%", 60, 0 },
 	[BTS_STAT_OML_CONNECTED] = \
 		{ "oml_connected",
 		  "Number of OML links connected",
 		  "", 16, 0   },
 	[BTS_STAT_RSL_CONNECTED] = \
 		{ "rsl_connected",
-		  "Number of RSL links connected",
+		  "Number of RSL links connected (same as num_trx:rsl_connected)",
 		  "", 16, 0   },
 	[BTS_STAT_LCHAN_BORKEN] = \
 		{ "lchan_borken",
@@ -1225,6 +1482,14 @@ const struct osmo_stat_item_desc bts_stat_desc[] = {
 		{ "ts_borken",
 		  "Number of timeslots in the BORKEN state",
 		  "", 16, 0 },
+	[BTS_STAT_NUM_TRX_RSL_CONNECTED] = \
+		{ "num_trx:rsl_connected",
+		  "Number of TRX in this BTS where RSL is up",
+		  "" },
+	[BTS_STAT_NUM_TRX_TOTAL] = \
+		{ "num_trx:total",
+		  "Number of configured TRX in this BTS",
+		  "" },
 };
 
 const struct osmo_stat_item_group_desc bts_statg_desc = {
