@@ -172,7 +172,7 @@ void gsm_mo_init(struct gsm_abis_mo *mo, struct gsm_bts *bts,
 const struct value_string gsm_chreq_descs[] = {
 	{ GSM_CHREQ_REASON_EMERG,	"emergency call" },
 	{ GSM_CHREQ_REASON_PAG,		"answer to paging" },
-	{ GSM_CHREQ_REASON_CALL,	"call re-establishment" },
+	{ GSM_CHREQ_REASON_CALL,	"call (re-)establishment" },
 	{ GSM_CHREQ_REASON_LOCATION_UPD,"Location updating" },
 	{ GSM_CHREQ_REASON_PDCH,	"one phase packet access" },
 	{ GSM_CHREQ_REASON_OTHER,	"other" },
@@ -191,7 +191,7 @@ const struct value_string gsm_pchant_names[] = {
 	{ GSM_PCHAN_UNKNOWN,	"UNKNOWN" },
 	{ GSM_PCHAN_CCCH_SDCCH4_CBCH, "CCCH+SDCCH4+CBCH" },
 	{ GSM_PCHAN_SDCCH8_SACCH8C_CBCH, "SDCCH8+CBCH" },
-	{ GSM_PCHAN_TCH_F_TCH_H_PDCH, "TCH/F_TCH/H_PDCH" },
+	{ GSM_PCHAN_OSMO_DYN, "TCH/F_TCH/H_SDCCH8_PDCH" },
 	{ 0,			NULL }
 };
 
@@ -207,7 +207,7 @@ const struct value_string gsm_pchan_ids[] = {
 	{ GSM_PCHAN_UNKNOWN,	"UNKNOWN" },
 	{ GSM_PCHAN_CCCH_SDCCH4_CBCH, "CCCH_SDCCH4_CBCH" },
 	{ GSM_PCHAN_SDCCH8_SACCH8C_CBCH, "SDCCH8_CBCH" },
-	{ GSM_PCHAN_TCH_F_TCH_H_PDCH, "TCH_F_TCH_H_PDCH" },
+	{ GSM_PCHAN_OSMO_DYN, "OSMO_DYN" },
 	{ 0,			NULL }
 };
 
@@ -224,7 +224,7 @@ const struct value_string gsm_pchant_descs[13] = {
 	{ GSM_PCHAN_UNKNOWN,	"Unknown / Unsupported channel combination" },
 	{ GSM_PCHAN_CCCH_SDCCH4_CBCH, "FCCH + SCH + BCCH + CCCH + CBCH + 3 SDCCH + 2 SACCH (Comb. V)" },
 	{ GSM_PCHAN_SDCCH8_SACCH8C_CBCH, "7 SDCCH + 4 SACCH + CBCH (Comb. VII)" },
-	{ GSM_PCHAN_TCH_F_TCH_H_PDCH, "Dynamic TCH/F or TCH/H or GPRS PDCH" },
+	{ GSM_PCHAN_OSMO_DYN, "Dynamic TCH/F or TCH/H or SDCCH/8 or GPRS PDCH" },
 	{ 0,			NULL }
 };
 
@@ -334,14 +334,15 @@ char *gsm_ts_and_pchan_name(const struct gsm_bts_trx_ts *ts)
 	return ts2str;
 }
 
-char *gsm_lchan_name_compute(const struct gsm_lchan *lchan)
+void lchan_update_name(struct gsm_lchan *lchan)
 {
 	struct gsm_bts_trx_ts *ts = lchan->ts;
-
-	snprintf(ts2str, sizeof(ts2str), "(bts=%d,trx=%d,ts=%d,ss=%d)",
-		 ts->trx->bts->nr, ts->trx->nr, ts->nr, lchan->nr);
-
-	return ts2str;
+	if (lchan->name)
+		talloc_free(lchan->name);
+	lchan->name = talloc_asprintf(ts->trx, "(bts=%d,trx=%d,ts=%d,ss=%s%d)",
+				      ts->trx->bts->nr, ts->trx->nr, ts->nr,
+				      lchan->vamos.is_secondary ? "shadow" : "",
+				      lchan->nr - (lchan->vamos.is_secondary ? ts->max_primary_lchans : 0));
 }
 
 /* obtain the MO structure for a given object instance */
@@ -494,25 +495,35 @@ gsm_objclass2obj(struct gsm_bts *bts, uint8_t obj_class,
 }
 
 /* See Table 10.5.25 of GSM04.08 */
-uint8_t gsm_pchan2chan_nr(enum gsm_phys_chan_config pchan,
-			  uint8_t ts_nr, uint8_t lchan_nr)
+int gsm_pchan2chan_nr(enum gsm_phys_chan_config pchan,
+		      uint8_t ts_nr, uint8_t lchan_nr, bool vamos_is_secondary)
 {
 	uint8_t cbits, chan_nr;
 
 	switch (pchan) {
 	case GSM_PCHAN_TCH_F:
 	case GSM_PCHAN_TCH_F_PDCH:
-		OSMO_ASSERT(lchan_nr == 0);
-		cbits = 0x01;
+		if (lchan_nr != 0)
+			return -EINVAL;
+		if (vamos_is_secondary)
+			cbits = ABIS_RSL_CHAN_NR_CBITS_OSMO_VAMOS_Bm_ACCHs;
+		else
+			cbits = 0x01;
 		break;
 	case GSM_PCHAN_PDCH:
-		OSMO_ASSERT(lchan_nr == 0);
+		if (lchan_nr != 0)
+			return -EINVAL;
 		cbits = RSL_CHAN_OSMO_PDCH >> 3;
 		break;
 	case GSM_PCHAN_TCH_H:
-		OSMO_ASSERT(lchan_nr < 2);
-		cbits = 0x02;
-		cbits += lchan_nr;
+		if (lchan_nr >= 2)
+			return -EINVAL;
+		if (vamos_is_secondary)
+			cbits = ABIS_RSL_CHAN_NR_CBITS_OSMO_VAMOS_Lm_ACCHs(lchan_nr);
+		else {
+			cbits = 0x02;
+			cbits += lchan_nr;
+		}
 		break;
 	case GSM_PCHAN_CCCH_SDCCH4:
 	case GSM_PCHAN_CCCH_SDCCH4_CBCH:
@@ -523,20 +534,22 @@ uint8_t gsm_pchan2chan_nr(enum gsm_phys_chan_config pchan,
 		 */
 		if (lchan_nr == CCCH_LCHAN)
 			chan_nr = 0;
-		else
-			OSMO_ASSERT(lchan_nr < 4);
+		else if (lchan_nr >= 4)
+			return -EINVAL;
 		cbits = 0x04;
 		cbits += lchan_nr;
 		break;
 	case GSM_PCHAN_SDCCH8_SACCH8C:
 	case GSM_PCHAN_SDCCH8_SACCH8C_CBCH:
-		OSMO_ASSERT(lchan_nr < 8);
+		if (lchan_nr >= 8)
+			return -EINVAL;
 		cbits = 0x08;
 		cbits += lchan_nr;
 		break;
 	default:
 	case GSM_PCHAN_CCCH:
-		OSMO_ASSERT(lchan_nr == 0);
+		if (lchan_nr != 0)
+			return -EINVAL;
 		cbits = 0x10;
 		break;
 	}
@@ -546,11 +559,43 @@ uint8_t gsm_pchan2chan_nr(enum gsm_phys_chan_config pchan,
 	return chan_nr;
 }
 
-uint8_t gsm_lchan2chan_nr(const struct gsm_lchan *lchan)
+/* For RSL, to talk to osmo-bts, we introduce Osmocom specific channel number cbits to indicate VAMOS secondary lchans.
+ * However, in RR, which is sent to the MS, these special cbits must not be sent, but their "normal" equivalent; for RR
+ * messages, pass allow_osmo_cbits = false. */
+int gsm_lchan_and_pchan2chan_nr(const struct gsm_lchan *lchan, enum gsm_phys_chan_config pchan, bool allow_osmo_cbits)
 {
-	/* Note: non-standard Osmocom style dyn TS PDCH mode chan_nr is only used within
-	 * rsl_tx_dyn_ts_pdch_act_deact(). */
-	return gsm_pchan2chan_nr(lchan->ts->pchan_is, lchan->ts->nr, lchan->nr);
+	int rc;
+	uint8_t lchan_nr = lchan->nr;
+
+	/* Take care that we never send Osmocom specific cbits to non-Osmo BTS. */
+	if (allow_osmo_cbits && lchan->vamos.is_secondary
+	    && lchan->ts->trx->bts->model->type != GSM_BTS_TYPE_OSMOBTS) {
+		LOG_LCHAN(lchan, LOGL_ERROR, "Cannot address VAMOS shadow lchan on this BTS type: %s\n",
+			  get_value_string(bts_type_names, lchan->ts->trx->bts->model->type));
+		return -ENOTSUP;
+	}
+	if (allow_osmo_cbits && lchan->ts->trx->bts->model->type != GSM_BTS_TYPE_OSMOBTS)
+		allow_osmo_cbits = false;
+
+	/* The VAMOS lchans are behind the primary ones in the ts->lchan[] array. They keep their lchan->nr as in the
+	 * array, but on the wire they are the "shadow" lchans for the primary lchans. For example, for TCH/F, there is
+	 * a primary ts->lchan[0] and a VAMOS ts->lchan[1]. Still, the VAMOS lchan should send chan_nr = 0. */
+	if (lchan->vamos.is_secondary)
+		lchan_nr -= lchan->ts->max_primary_lchans;
+	rc = gsm_pchan2chan_nr(pchan, lchan->ts->nr, lchan_nr,
+			       allow_osmo_cbits ? lchan->vamos.is_secondary : false);
+	/* Log an error so that we don't need to add logging to each caller of this function */
+	if (rc < 0)
+		LOG_LCHAN(lchan, LOGL_ERROR,
+			  "Error encoding Channel Number: pchan %s ts %u ss %u%s\n",
+			  gsm_pchan_name(lchan->ts->pchan_from_config), lchan->ts->nr, lchan->nr,
+			  lchan->vamos.is_secondary ? " (VAMOS shadow)" : "");
+	return rc;
+}
+
+int gsm_lchan2chan_nr(const struct gsm_lchan *lchan, bool allow_osmo_cbits)
+{
+	return gsm_lchan_and_pchan2chan_nr(lchan, lchan->ts->pchan_is, allow_osmo_cbits);
 }
 
 static const uint8_t subslots_per_pchan[] = {
@@ -564,12 +609,12 @@ static const uint8_t subslots_per_pchan[] = {
 	[GSM_PCHAN_CCCH_SDCCH4_CBCH] = 4,
 	[GSM_PCHAN_SDCCH8_SACCH8C_CBCH] = 8,
 	/* Dyn TS: maximum allowed subslots */
-	[GSM_PCHAN_TCH_F_TCH_H_PDCH] = 2,
+	[GSM_PCHAN_OSMO_DYN] = 8,
 	[GSM_PCHAN_TCH_F_PDCH] = 1,
 };
 
-/*! According to ts->pchan and possibly ts->dyn_pchan, return the number of
- * logical channels available in the timeslot. */
+/*! Return the maximum number of logical channels that may be used in a timeslot of the given physical channel
+ * configuration. */
 uint8_t pchan_subslots(enum gsm_phys_chan_config pchan)
 {
 	if (pchan < 0 || pchan >= ARRAY_SIZE(subslots_per_pchan))
@@ -577,6 +622,30 @@ uint8_t pchan_subslots(enum gsm_phys_chan_config pchan)
 	return subslots_per_pchan[pchan];
 }
 
+static const uint8_t subslots_per_pchan_vamos[] = {
+	[GSM_PCHAN_NONE] = 0,
+	[GSM_PCHAN_CCCH] = 0,
+	[GSM_PCHAN_PDCH] = 0,
+	[GSM_PCHAN_CCCH_SDCCH4] = 0,
+	/* VAMOS: on a TCH/F, there may be a TCH/H shadow */
+	[GSM_PCHAN_TCH_F] = 2,
+	[GSM_PCHAN_TCH_H] = 2,
+	[GSM_PCHAN_SDCCH8_SACCH8C] = 0,
+	[GSM_PCHAN_CCCH_SDCCH4_CBCH] = 0,
+	[GSM_PCHAN_SDCCH8_SACCH8C_CBCH] = 0,
+	[GSM_PCHAN_OSMO_DYN] = 0,
+	[GSM_PCHAN_TCH_F_PDCH] = 2,
+};
+
+/* Return the maximum number of VAMOS secondary lchans that may be used in a timeslot of the given physical channel
+ * configuration. */
+uint8_t pchan_subslots_vamos(enum gsm_phys_chan_config pchan)
+{
+	if (pchan < 0 || pchan >= ARRAY_SIZE(subslots_per_pchan_vamos))
+		return 0;
+	return subslots_per_pchan_vamos[pchan];
+}
+
 static bool pchan_is_tch(enum gsm_phys_chan_config pchan)
 {
 	switch (pchan) {
@@ -593,23 +662,54 @@ bool ts_is_tch(struct gsm_bts_trx_ts *ts)
 	return pchan_is_tch(ts->pchan_is);
 }
 
+/* For a VAMOS secondary shadow lchan, return its primary lchan. If the lchan is not a secondary lchan, return NULL. */
+struct gsm_lchan *gsm_lchan_vamos_to_primary(const struct gsm_lchan *lchan_vamos)
+{
+	struct gsm_lchan *lchan_primary;
+	if (!lchan_vamos || !lchan_vamos->vamos.is_secondary)
+		return NULL;
+	/* OsmoBSC currently does not support mixed TCH/F + TCH/H VAMOS multiplexes. Hence the primary <-> secondary
+	 * relation is a simple index shift in the lchan array. If mixed multiplexes were allowed, a TCH/F primary might
+	 * have two TCH/H VAMOS secondary lchans, etc. Fortunately, we don't need to care about that. */
+	lchan_primary = (struct gsm_lchan*)lchan_vamos - lchan_vamos->ts->max_primary_lchans;
+	if (!lchan_primary->fi)
+		return NULL;
+	return lchan_primary;
+}
+
+/* For a primary lchan, return its VAMOS secondary shadow lchan. If the lchan is not a primary lchan, return NULL. */
+struct gsm_lchan *gsm_lchan_primary_to_vamos(const struct gsm_lchan *lchan_primary)
+{
+	struct gsm_lchan *lchan_vamos;
+	if (!lchan_primary || lchan_primary->vamos.is_secondary)
+		return NULL;
+	/* OsmoBSC currently does not support mixed TCH/F + TCH/H VAMOS multiplexes. Hence the primary <-> secondary
+	 * relation is a simple index shift in the lchan array. If mixed multiplexes were allowed, a TCH/F primary might
+	 * have two TCH/H VAMOS secondary lchans, etc. Fortunately, we don't need to care about that. */
+	lchan_vamos = (struct gsm_lchan*)lchan_primary + lchan_primary->ts->max_primary_lchans;
+	if (!lchan_vamos->fi)
+		return NULL;
+	return lchan_vamos;
+}
+
 struct gsm_bts *conn_get_bts(struct gsm_subscriber_connection *conn) {
 	if (!conn || !conn->lchan)
 		return NULL;
 	return conn->lchan->ts->trx->bts;
 }
 
-static void _chan_desc_fill_tail(struct gsm48_chan_desc *cd, const struct gsm_lchan *lchan)
+static void _chan_desc_fill_tail(struct gsm48_chan_desc *cd, const struct gsm_lchan *lchan,
+				 uint8_t tsc)
 {
 	if (!lchan->ts->hopping.enabled) {
 		uint16_t arfcn = lchan->ts->trx->arfcn & 0x3ff;
-		cd->h0.tsc = gsm_ts_tsc(lchan->ts);
+		cd->h0.tsc = tsc;
 		cd->h0.h = 0;
 		cd->h0.spare = 0;
 		cd->h0.arfcn_high = arfcn >> 8;
 		cd->h0.arfcn_low = arfcn & 0xff;
 	} else {
-		cd->h1.tsc = gsm_ts_tsc(lchan->ts);
+		cd->h1.tsc = tsc;
 		cd->h1.h = 1;
 		cd->h1.maio_high = lchan->ts->hopping.maio >> 2;
 		cd->h1.maio_low = lchan->ts->hopping.maio & 0x03;
@@ -617,21 +717,30 @@ static void _chan_desc_fill_tail(struct gsm48_chan_desc *cd, const struct gsm_lc
 	}
 }
 
-void gsm48_lchan2chan_desc(struct gsm48_chan_desc *cd,
-			   const struct gsm_lchan *lchan)
+int gsm48_lchan_and_pchan2chan_desc(struct gsm48_chan_desc *cd,
+				    const struct gsm_lchan *lchan,
+				    enum gsm_phys_chan_config pchan,
+				    uint8_t tsc, bool allow_osmo_cbits)
 {
-	cd->chan_nr = gsm_lchan2chan_nr(lchan);
-	_chan_desc_fill_tail(cd, lchan);
+	int chan_nr = gsm_lchan_and_pchan2chan_nr(lchan, pchan, allow_osmo_cbits);
+	if (chan_nr < 0) {
+		/* Log an error so that we don't need to add logging to each caller of this function */
+		LOG_LCHAN(lchan, LOGL_ERROR,
+			  "Error encoding Channel Number: pchan %s ts %u ss %u%s (rc = %d)\n",
+			  gsm_pchan_name(pchan), lchan->ts->nr, lchan->nr,
+			  lchan->vamos.is_secondary ? " (VAMOS shadow)" : "", chan_nr);
+		return chan_nr;
+	}
+	cd->chan_nr = chan_nr;
+	_chan_desc_fill_tail(cd, lchan, tsc);
+	return 0;
 }
 
-/* like gsm48_lchan2chan_desc() above, but use ts->pchan_from_config to
- * return a channel description based on what is configured, rather than
- * what the current state of the pchan type is */
-void gsm48_lchan2chan_desc_as_configured(struct gsm48_chan_desc *cd,
-					 const struct gsm_lchan *lchan)
+int gsm48_lchan2chan_desc(struct gsm48_chan_desc *cd,
+			  const struct gsm_lchan *lchan,
+			  uint8_t tsc, bool allow_osmo_cbits)
 {
-	cd->chan_nr = gsm_pchan2chan_nr(lchan->ts->pchan_from_config, lchan->ts->nr, lchan->nr);
-	_chan_desc_fill_tail(cd, lchan);
+	return gsm48_lchan_and_pchan2chan_desc(cd, lchan, lchan->ts->pchan_is, tsc, allow_osmo_cbits);
 }
 
 uint8_t gsm_ts_tsc(const struct gsm_bts_trx_ts *ts)
@@ -677,6 +786,8 @@ enum gsm_phys_chan_config gsm_pchan_by_lchan_type(enum gsm_chan_t type)
 		return GSM_PCHAN_TCH_F;
 	case GSM_LCHAN_TCH_H:
 		return GSM_PCHAN_TCH_H;
+	case GSM_LCHAN_SDCCH:
+		return GSM_PCHAN_SDCCH8_SACCH8C;
 	case GSM_LCHAN_NONE:
 	case GSM_LCHAN_PDTCH:
 		/* TODO: so far lchan->type is NONE in PDCH mode. PDTCH is only
@@ -688,6 +799,22 @@ enum gsm_phys_chan_config gsm_pchan_by_lchan_type(enum gsm_chan_t type)
 	}
 }
 
+enum channel_rate chan_t_to_chan_rate(enum gsm_chan_t chan_t)
+{
+	switch (chan_t) {
+	case GSM_LCHAN_SDCCH:
+		return CH_RATE_SDCCH;
+	case GSM_LCHAN_TCH_F:
+		return CH_RATE_FULL;
+	case GSM_LCHAN_TCH_H:
+		return CH_RATE_HALF;
+	default:
+		/* For other channel types, the channel_rate value is never used. It is fine to return an invalid value,
+		 * and callers don't actually need to check for this. */
+		return -1;
+	}
+}
+
 /* Can the timeslot in principle be used as this PCHAN kind? */
 bool ts_is_capable_of_pchan(struct gsm_bts_trx_ts *ts, enum gsm_phys_chan_config pchan)
 {
@@ -701,11 +828,12 @@ bool ts_is_capable_of_pchan(struct gsm_bts_trx_ts *ts, enum gsm_phys_chan_config
 			return false;
 		}
 
-	case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+	case GSM_PCHAN_OSMO_DYN:
 		switch (pchan) {
 		case GSM_PCHAN_TCH_F:
 		case GSM_PCHAN_TCH_H:
 		case GSM_PCHAN_PDCH:
+		case GSM_PCHAN_SDCCH8_SACCH8C:
 			return true;
 		default:
 			return false;
@@ -773,11 +901,12 @@ bool ts_is_capable_of_lchant(struct gsm_bts_trx_ts *ts, enum gsm_chan_t type)
 			return false;
 		}
 
-	case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+	case GSM_PCHAN_OSMO_DYN:
 		switch (type) {
 		case GSM_LCHAN_TCH_F:
 		case GSM_LCHAN_TCH_H:
 		case GSM_LCHAN_PDTCH:
+		case GSM_LCHAN_SDCCH:
 			return true;
 		default:
 			return false;
@@ -911,11 +1040,26 @@ ms_power_default:
 }
 
 const struct value_string lchan_activate_mode_names[] = {
-	OSMO_VALUE_STRING(FOR_NONE),
-	OSMO_VALUE_STRING(FOR_MS_CHANNEL_REQUEST),
-	OSMO_VALUE_STRING(FOR_ASSIGNMENT),
-	OSMO_VALUE_STRING(FOR_HANDOVER),
-	OSMO_VALUE_STRING(FOR_VTY),
+	OSMO_VALUE_STRING(ACTIVATE_FOR_NONE),
+	OSMO_VALUE_STRING(ACTIVATE_FOR_MS_CHANNEL_REQUEST),
+	OSMO_VALUE_STRING(ACTIVATE_FOR_ASSIGNMENT),
+	OSMO_VALUE_STRING(ACTIVATE_FOR_HANDOVER),
+	OSMO_VALUE_STRING(ACTIVATE_FOR_VTY),
+	{}
+};
+
+const struct value_string lchan_modify_for_names[] = {
+	OSMO_VALUE_STRING(MODIFY_FOR_NONE),
+	OSMO_VALUE_STRING(MODIFY_FOR_ASSIGNMENT),
+	OSMO_VALUE_STRING(MODIFY_FOR_VTY),
+	{}
+};
+
+const struct value_string assign_for_names[] = {
+	OSMO_VALUE_STRING(ASSIGN_FOR_NONE),
+	OSMO_VALUE_STRING(ASSIGN_FOR_BSSMAP_REQ),
+	OSMO_VALUE_STRING(ASSIGN_FOR_CONGESTION_RESOLUTION),
+	OSMO_VALUE_STRING(ASSIGN_FOR_VTY),
 	{}
 };
 
@@ -972,6 +1116,19 @@ enum gsm48_rr_cause bsc_gsm48_rr_cause_from_rsl_cause(uint8_t c)
 	}
 }
 
+/* Default Interference Measurement Parameters */
+const struct gsm_interf_meas_params interf_meas_params_def = {
+	.avg_period = 6, /* 6 SACCH periods */
+	.bounds_dbm = {
+		115, /*  0: -115 dBm */
+		109, /* X1: -109 dBm */
+		103, /* X2: -103 dBm */
+		 97, /* X3:  -97 dBm */
+		 91, /* X4:  -91 dBm */
+		 85, /* X5:  -85 dBm */
+	},
+};
+
 /* Default MS/BS Power Control parameters (see 3GPP TS 45.008, table A.1) */
 const struct gsm_power_ctrl_params power_ctrl_params_def = {
 	/* Static Power Control is the safe default */
@@ -985,8 +1142,11 @@ const struct gsm_power_ctrl_params power_ctrl_params_def = {
 	.inc_step_size_db = 4, /* 2, 4, or 6 dB */
 	.red_step_size_db = 2, /* 2 or 4 dB */
 
+	.ctrl_interval = 1, /* Trigger loop every second SACCH block. TS 45.008 sec 4.7.1 */
+
 	/* RxLev measurement parameters */
 	.rxlev_meas = {
+		.enabled = true,
 		/* Thresholds for RxLev (see 3GPP TS 45.008, A.3.2.1) */
 		.lower_thresh = 32, /* L_RXLEV_XX_P (-78 dBm) */
 		.upper_thresh = 38, /* U_RXLEV_XX_P (-72 dBm) */
@@ -1011,16 +1171,155 @@ const struct gsm_power_ctrl_params power_ctrl_params_def = {
 
 	/* RxQual measurement parameters */
 	.rxqual_meas = {
+		.enabled = true,
 		/* Thresholds for RxQual (see 3GPP TS 45.008, A.3.2.1) */
 		.lower_thresh = 3, /* L_RXQUAL_XX_P (0.8% <= BER < 1.6%) */
 		.upper_thresh = 0, /* U_RXQUAL_XX_P (BER < 0.2%) */
 
 		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
-		 * out of LOWER_CMP_N averages are lower than L_RXLEV_XX_P */
+		 * out of LOWER_CMP_N averages are lower than L_RXQUAL_XX_P */
 		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
 		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
 		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
-		 * out of UPPER_CMP_N averages are greater than L_RXLEV_XX_P */
+		 * out of UPPER_CMP_N averages are greater than L_RXQUAL_XX_P */
+		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+
+		/* No averaging (filtering) by default */
+		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
+
+		/* Hreqave: the period over which an average is produced */
+		.h_reqave = 4, /* TODO: investigate a reasonable default value */
+		/* Hreqt: the number of averaged results maintained */
+		.h_reqt = 6, /* TODO: investigate a reasonable default value */
+	},
+
+	/* C/I measurement parameters.
+	 * Target C/I retrieved from "GSM/EDGE: Evolution and Performance" Table 10.3.
+	 * Set lower and upper so that (lower + upper) / 2 is equal or slightly
+	 * above the target.
+	 */
+	.ci_fr_meas = { /* FR: Target C/I = 15 dB, Soft blocking threshold = 10 dB */
+		.enabled = true,
+		.lower_thresh = 13,
+		.upper_thresh = 17,
+
+		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
+		 * out of LOWER_CMP_N averages are lower than L_CI_FR_XX_P */
+		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
+		 * out of UPPER_CMP_N averages are greater than L_CI_FR_XX_P */
+		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+
+		/* No averaging (filtering) by default */
+		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
+
+		/* Hreqave: the period over which an average is produced */
+		.h_reqave = 4, /* TODO: investigate a reasonable default value */
+		/* Hreqt: the number of averaged results maintained */
+		.h_reqt = 6, /* TODO: investigate a reasonable default value */
+	},
+	.ci_hr_meas = { /* HR: Target C/I = 18 dB, Soft blocking threshold = 13 dB */
+		.enabled = true,
+		.lower_thresh = 16,
+		.upper_thresh = 21,
+
+		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
+		 * out of LOWER_CMP_N averages are lower than L_CI_HR_XX_P */
+		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
+		 * out of UPPER_CMP_N averages are greater than L_CI_HR_XX_P */
+		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+
+		/* No averaging (filtering) by default */
+		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
+
+		/* Hreqave: the period over which an average is produced */
+		.h_reqave = 4, /* TODO: investigate a reasonable default value */
+		/* Hreqt: the number of averaged results maintained */
+		.h_reqt = 6, /* TODO: investigate a reasonable default value */
+	},
+	.ci_amr_fr_meas = { /* AMR-FR: Target C/I = 9 dB, Soft blocking threshold = 4 dB */
+		.enabled = true,
+		.lower_thresh = 7,
+		.upper_thresh = 11,
+
+		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
+		 * out of LOWER_CMP_N averages are lower than L_CI_AMR_FR_XX_P */
+		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
+		 * out of UPPER_CMP_N averages are greater than L_CI_AMR_FR_XX_P */
+		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+
+		/* No averaging (filtering) by default */
+		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
+
+		/* Hreqave: the period over which an average is produced */
+		.h_reqave = 4, /* TODO: investigate a reasonable default value */
+		/* Hreqt: the number of averaged results maintained */
+		.h_reqt = 6, /* TODO: investigate a reasonable default value */
+	},
+	.ci_amr_hr_meas = { /* AMR-HR: Target C/I = 15 dB, Soft blocking threshold = 10 dB */
+		.enabled = true,
+		.lower_thresh = 13,
+		.upper_thresh = 17,
+
+		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
+		 * out of LOWER_CMP_N averages are lower than L_CI_AMR_HR_XX_P */
+		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
+		 * out of UPPER_CMP_N averages are greater than L_CI_AMR_HR_XX_P */
+		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+
+		/* No averaging (filtering) by default */
+		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
+
+		/* Hreqave: the period over which an average is produced */
+		.h_reqave = 4, /* TODO: investigate a reasonable default value */
+		/* Hreqt: the number of averaged results maintained */
+		.h_reqt = 6, /* TODO: investigate a reasonable default value */
+	},
+	.ci_sdcch_meas = { /* SDCCH: Target C/I = 14 dB, Soft blocking threshold = 9 dB */
+		.enabled = true,
+		.lower_thresh = 12,
+		.upper_thresh = 16,
+
+		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
+		 * out of LOWER_CMP_N averages are lower than L_CI_SDCCH_XX_P */
+		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
+		 * out of UPPER_CMP_N averages are greater than L_CI_SDCCH_XX_P */
+		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
+
+		/* No averaging (filtering) by default */
+		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
+
+		/* Hreqave: the period over which an average is produced */
+		.h_reqave = 4, /* TODO: investigate a reasonable default value */
+		/* Hreqt: the number of averaged results maintained */
+		.h_reqt = 6, /* TODO: investigate a reasonable default value */
+	},
+	.ci_gprs_meas = { /* GPRS: Target C/I = 20 dB, Soft blocking threshold = 15 dB */
+		.enabled = true,
+		.lower_thresh = 18,
+		.upper_thresh = 24,
+
+		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
+		 * out of LOWER_CMP_N averages are lower than L_CI_GPRS_XX_P */
+		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
+		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
+		 * out of UPPER_CMP_N averages are greater than L_CI_GPRS_XX_P */
 		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
 		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
 
@@ -1033,3 +1332,32 @@ const struct gsm_power_ctrl_params power_ctrl_params_def = {
 		.h_reqt = 6, /* TODO: investigate a reasonable default value */
 	},
 };
+
+void power_ctrl_params_def_reset(struct gsm_power_ctrl_params *params,
+				 enum gsm_power_ctrl_dir dir)
+{
+	*params = power_ctrl_params_def;
+	params->dir = dir;
+	if (dir == GSM_PWR_CTRL_DIR_UL)
+		/* Trigger loop every fourth SACCH block (1.92s). TS 45.008 sec 4.7.1: */
+		params->ctrl_interval = 2;
+}
+
+enum rsl_cmod_spd chan_mode_to_rsl_cmod_spd(enum gsm48_chan_mode chan_mode)
+{
+	switch (gsm48_chan_mode_to_non_vamos(chan_mode)) {
+	case GSM48_CMODE_SIGN:
+		return RSL_CMOD_SPD_SIGN;
+	case GSM48_CMODE_SPEECH_V1:
+	case GSM48_CMODE_SPEECH_EFR:
+	case GSM48_CMODE_SPEECH_AMR:
+		return RSL_CMOD_SPD_SPEECH;
+	case GSM48_CMODE_DATA_14k5:
+	case GSM48_CMODE_DATA_12k0:
+	case GSM48_CMODE_DATA_6k0:
+	case GSM48_CMODE_DATA_3k6:
+		return RSL_CMOD_SPD_DATA;
+	default:
+		return -EINVAL;
+	}
+}
