@@ -29,6 +29,7 @@
 
 #include <osmocom/bsc/debug.h>
 #include <osmocom/bsc/gsm_data.h>
+#include <osmocom/bsc/assignment_fsm.h>
 #include <osmocom/bsc/handover_fsm.h>
 #include <osmocom/bsc/handover_decision.h>
 #include <osmocom/bsc/handover_decision_2.h>
@@ -41,6 +42,7 @@
 #include <osmocom/bsc/timeslot_fsm.h>
 #include <osmocom/bsc/bts.h>
 #include <osmocom/bsc/lchan_select.h>
+#include <osmocom/bsc/chan_counts.h>
 
 #define LOGPHOBTS(bts, level, fmt, args...) \
 	LOGP(DHODEC, level, "(BTS %u) " fmt, bts->nr, ## args)
@@ -52,7 +54,7 @@
 	     lchan->ts->nr, \
 	     lchan->nr, \
 	     gsm_lchant_name(lchan->type), \
-	     gsm48_chan_mode_name(lchan->tch_mode), \
+	     gsm48_chan_mode_name(lchan->current_ch_mode_rate.chan_mode), \
 	     bsc_subscr_name(lchan->conn? lchan->conn->bsub : NULL), \
 	     ## args)
 
@@ -63,7 +65,7 @@
 	     lchan->ts->nr, \
 	     lchan->nr, \
 	     gsm_lchant_name(lchan->type), \
-	     gsm48_chan_mode_name(lchan->tch_mode), \
+	     gsm48_chan_mode_name(lchan->current_ch_mode_rate.chan_mode), \
 	     new_bts->nr, \
 	     bsc_subscr_name(lchan->conn? lchan->conn->bsub : NULL), \
 	     ## args)
@@ -75,7 +77,7 @@
 	     lchan->ts->nr, \
 	     lchan->nr, \
 	     gsm_lchant_name(lchan->type), \
-	     gsm48_chan_mode_name(lchan->tch_mode), \
+	     gsm48_chan_mode_name(lchan->current_ch_mode_rate.chan_mode), \
 	     gsm0808_cell_id_list_name(remote_cil), \
 	     bsc_subscr_name(lchan->conn? lchan->conn->bsub : NULL), \
 	     ## args)
@@ -83,9 +85,9 @@
 #define LOGPHOCAND(candidate, level, fmt, args...) do {\
 	if ((candidate)->target.bts) \
 		LOGPHOLCHANTOBTS((candidate)->current.lchan, (candidate)->target.bts, level, fmt, ## args); \
-	else if ((candidate)->target.cil) \
-		LOGPHOLCHANTOREMOTE((candidate)->current.lchan, (candidate)->target.cil, level, fmt, ## args); \
-	} while(0)
+	else if ((candidate)->target.cell_ids.id_list_len) \
+		LOGPHOLCHANTOREMOTE((candidate)->current.lchan, &(candidate)->target.cell_ids, level, fmt, ## args); \
+	} while (0)
 
 
 #define REQUIREMENT_A_TCHF	0x01
@@ -120,8 +122,8 @@ struct ho_candidate {
 		int lchan_frees_tchh;
 	} current;
 	struct {
-		struct neighbor_ident_key nik;	/* neighbor ARFCN+BSIC */
-		const struct gsm0808_cell_id_list2 *cil; /* target cells in remote BSS */
+		struct cell_ab ab;	/* neighbor ARFCN+BSIC */
+		struct gsm0808_cell_id_list2 cell_ids; /* target cells in remote BSS */
 		struct gsm_bts *bts;
 		int rxlev;
 		int rxlev_afs_bias;
@@ -166,11 +168,18 @@ static enum ho_reason global_ho_reason;
 
 static void congestion_check_cb(void *arg);
 
+static bool is_upgrade_to_tchf(const struct ho_candidate *c, uint8_t for_requirement)
+{
+	return c->current.lchan
+		&& (c->current.lchan->type == GSM_LCHAN_TCH_H)
+		&& ((c->requirements & for_requirement) & REQUIREMENT_TCHF_MASK);
+}
+
 static unsigned int ts_usage_count(struct gsm_bts_trx_ts *ts)
 {
 	struct gsm_lchan *lchan;
 	unsigned int count = 0;
-	ts_for_each_lchan(lchan, ts) {
+	ts_for_n_lchans(lchan, ts, ts->max_lchans_possible) {
 		if (lchan_state_is(lchan, LCHAN_ST_ESTABLISHED))
 			count++;
 	}
@@ -215,49 +224,6 @@ void hodec2_on_change_congestion_check_interval(struct gsm_network *net, unsigne
 	reinit_congestion_timer(net);
 }
 
-static void _conn_penalty_time_add(struct gsm_subscriber_connection *conn,
-				   const void *for_object,
-				   int penalty_time)
-{
-	if (!for_object) {
-		LOGP(DHODEC, LOGL_ERROR, "%s Unable to set Handover-2 penalty timer:"
-		     " no target cell pointer\n",
-		     bsc_subscr_name(conn->bsub));
-		return;
-	}
-
-	if (!conn->hodec2.penalty_timers) {
-		conn->hodec2.penalty_timers = penalty_timers_init(conn);
-		OSMO_ASSERT(conn->hodec2.penalty_timers);
-	}
-
-	penalty_timers_add(conn->hodec2.penalty_timers, for_object, penalty_time);
-}
-
-static void nik_penalty_time_add(struct gsm_subscriber_connection *conn,
-				 struct neighbor_ident_key *nik,
-				 int penalty_time)
-{
-	_conn_penalty_time_add(conn,
-			       neighbor_ident_get(conn->network->neighbor_bss_cells, nik),
-			       penalty_time);
-}
-
-static void bts_penalty_time_add(struct gsm_subscriber_connection *conn,
-				 struct gsm_bts *bts,
-				 int penalty_time)
-{
-	_conn_penalty_time_add(conn, bts, penalty_time);
-}
-
-static unsigned int conn_penalty_time_remaining(struct gsm_subscriber_connection *conn,
-						const void *for_object)
-{
-	if (!conn->hodec2.penalty_timers)
-		return 0;
-	return penalty_timers_remaining(conn->hodec2.penalty_timers, for_object);
-}
-
 /* did we get a RXLEV for a given cell in the given report? Mark matches as MRC_F_PROCESSED. */
 static struct gsm_meas_rep_cell *cell_in_rep(struct gsm_meas_rep *mr, uint16_t arfcn, uint8_t bsic)
 {
@@ -279,12 +245,32 @@ static struct gsm_meas_rep_cell *cell_in_rep(struct gsm_meas_rep *mr, uint16_t a
 static int current_rxlev(struct gsm_lchan *lchan)
 {
 	struct gsm_bts *bts = lchan->ts->trx->bts;
-	return get_meas_rep_avg(lchan,
-				ho_get_hodec2_full_tdma(bts->ho) ?
-					MEAS_REP_DL_RXLEV_FULL : MEAS_REP_DL_RXLEV_SUB,
+	return get_meas_rep_avg(lchan, TDMA_MEAS_FIELD_RXLEV, TDMA_MEAS_DIR_DL, ho_get_hodec2_tdma_meas_set(bts->ho),
 				ho_get_hodec2_rxlev_avg_win(bts->ho));
 }
 
+static int current_rxqual(struct gsm_lchan *lchan)
+{
+	struct gsm_bts *bts = lchan->ts->trx->bts;
+	return get_meas_rep_avg(lchan, TDMA_MEAS_FIELD_RXQUAL, TDMA_MEAS_DIR_DL, ho_get_hodec2_tdma_meas_set(bts->ho),
+				ho_get_hodec2_rxqual_avg_win(bts->ho));
+}
+
+static bool is_low_rxlev(int rxlev_current, struct handover_cfg *neigh_cfg)
+{
+	return rxlev_current >= 0
+		&& rxlev2dbm(rxlev_current) < ho_get_hodec2_min_rxlev(neigh_cfg);
+}
+
+static bool is_low_rxqual(int rxqual_current, struct handover_cfg *neigh_cfg)
+{
+	/* min_rxqual is actually a bit of a misnomer, low quality is a high number. So the "min" refers to the minimum
+	 * acceptable level of quality, and "min or better" here means "rxqual number must be SMALLER-or-equal than the
+	 * min-rxqual setting". */
+	return rxqual_current >= 0
+		&& rxqual_current > ho_get_hodec2_min_rxqual(neigh_cfg);
+}
+
 /* obtain averaged rxlev for given neighbor */
 static int neigh_meas_avg(struct neigh_meas_proc *nmp, int window)
 {
@@ -491,6 +477,7 @@ static void check_requirements(struct ho_candidate *c)
 	uint8_t requirement = 0;
 	unsigned int penalty_time;
 	int32_t current_overbooked;
+	struct gsm0808_cell_id target_cell_id;
 	c->requirements = 0;
 
 	/* Requirement A */
@@ -510,7 +497,8 @@ static void check_requirements(struct ho_candidate *c)
 	}
 
 	/* the handover penalty timer must not run for this bts */
-	penalty_time = conn_penalty_time_remaining(c->current.lchan->conn, c->target.bts);
+	gsm_bts_cell_id(&target_cell_id, c->target.bts);
+	penalty_time = penalty_timers_remaining(&c->current.lchan->conn->hodec2.penalty_timers, &target_cell_id);
 	if (penalty_time) {
 		LOGPHOLCHANTOBTS(c->current.lchan, c->target.bts, LOGL_DEBUG, "not a candidate, target BTS still in penalty time"
 				 " (%u seconds left)\n", penalty_time);
@@ -519,7 +507,7 @@ static void check_requirements(struct ho_candidate *c)
 
 	/* compatibility check for codecs.
 	 * if so, the candidates for full rate and half rate are selected */
-	switch (c->current.lchan->tch_mode) {
+	switch (gsm48_chan_mode_to_non_vamos(c->current.lchan->current_ch_mode_rate.chan_mode)) {
 	case GSM48_CMODE_SPEECH_V1:
 		switch (c->current.lchan->type) {
 		case GSM_LCHAN_TCH_F: /* mandatory */
@@ -530,7 +518,7 @@ static void check_requirements(struct ho_candidate *c)
 				LOGPHOLCHANTOBTS(c->current.lchan, c->target.bts, LOGL_DEBUG,
 						 "tch_mode='%s' type='%s' not supported\n",
 						 get_value_string(gsm48_chan_mode_names,
-								  c->current.lchan->tch_mode),
+								  c->current.lchan->current_ch_mode_rate.chan_mode),
 						 gsm_lchant_name(c->current.lchan->type));
 				break;
 			}
@@ -539,7 +527,8 @@ static void check_requirements(struct ho_candidate *c)
 			break;
 		default:
 			LOGPHOLCHAN(c->current.lchan, LOGL_ERROR, "Unexpected channel type: neither TCH/F nor TCH/H for %s\n",
-				    get_value_string(gsm48_chan_mode_names, c->current.lchan->tch_mode));
+				    get_value_string(gsm48_chan_mode_names,
+						     c->current.lchan->current_ch_mode_rate.chan_mode));
 			return;
 		}
 		break;
@@ -614,7 +603,7 @@ static void check_requirements(struct ho_candidate *c)
 	// This was useful in osmo-nitb. We're in osmo-bsc now and have no idea whether the osmo-msc does
 	// internal or external call control. Maybe a future config switch wants to add this behavior?
 	/* Built-in call control requires equal codec rates. Remove rates that are not equal. */
-	if (c->current.lchan->tch_mode == GSM48_CMODE_SPEECH_AMR
+	if (gsm48_chan_mode_to_non_vamos(c->current.lchan->tch_mode) == GSM48_CMODE_SPEECH_AMR
 	    && c->current.bts->network->mncc_recv != mncc_sock_from_cc) {
 		switch (c->current.lchan->type) {
 		case GSM_LCHAN_TCH_F:
@@ -788,9 +777,9 @@ static void check_requirements_remote_bss(struct ho_candidate *c)
 	/* Requirement A */
 
 	/* the handover penalty timer must not run for this bts */
-	penalty_time = conn_penalty_time_remaining(c->current.lchan->conn, c->target.cil);
+	penalty_time = penalty_timers_remaining_list(&c->current.lchan->conn->hodec2.penalty_timers, &c->target.cell_ids);
 	if (penalty_time) {
-		LOGPHOLCHANTOREMOTE(c->current.lchan, c->target.cil, LOGL_DEBUG,
+		LOGPHOLCHANTOREMOTE(c->current.lchan, &c->target.cell_ids, LOGL_DEBUG,
 				    "not a candidate, target BSS still in penalty time"
 				    " (%u seconds left)\n", penalty_time);
 		return;
@@ -798,7 +787,7 @@ static void check_requirements_remote_bss(struct ho_candidate *c)
 
 	/* compatibility check for codecs -- we have no notion of what the remote BSS supports. We can
 	 * only assume that a handover would work, and use only the local requirements. */
-	switch (c->current.lchan->tch_mode) {
+	switch (gsm48_chan_mode_to_non_vamos(c->current.lchan->current_ch_mode_rate.chan_mode)) {
 	case GSM48_CMODE_SPEECH_V1:
 		switch (c->current.lchan->type) {
 		case GSM_LCHAN_TCH_F: /* mandatory */
@@ -810,7 +799,8 @@ static void check_requirements_remote_bss(struct ho_candidate *c)
 			break;
 		default:
 			LOGPHOLCHAN(c->current.lchan, LOGL_ERROR, "Unexpected channel type: neither TCH/F nor TCH/H for %s\n",
-				    get_value_string(gsm48_chan_mode_names, c->current.lchan->tch_mode));
+				    get_value_string(gsm48_chan_mode_names,
+						     c->current.lchan->current_ch_mode_rate.chan_mode));
 			return;
 		}
 		break;
@@ -851,12 +841,11 @@ static void check_requirements_remote_bss(struct ho_candidate *c)
 /* Trigger handover or assignment depending on the target BTS */
 static int trigger_local_ho_or_as(struct ho_candidate *c, uint8_t requirements)
 {
-	struct handover_out_req req;
 	int afs_bias = 0;
 	bool full_rate = false;
 
 	/* afs_bias becomes > 0, if AFS is used and is improved */
-	if (c->current.lchan->tch_mode == GSM48_CMODE_SPEECH_AMR)
+	if (gsm48_chan_mode_to_non_vamos(c->current.lchan->current_ch_mode_rate.chan_mode) == GSM48_CMODE_SPEECH_AMR)
 		afs_bias = ho_get_hodec2_afs_bias_rxlev(c->target.bts->ho);
 
 	/* select TCH rate, prefer TCH/F if AFS is improved */
@@ -866,7 +855,7 @@ static int trigger_local_ho_or_as(struct ho_candidate *c, uint8_t requirements)
 		if ((requirements & REQUIREMENT_TCHF_MASK)) {
 			if (c->current.bts == c->target.bts) {
 				LOGPHOLCHAN(c->current.lchan, LOGL_INFO, "Not performing assignment: Already on target type\n");
-				return 0;
+				return -EALREADY;
 			}
 			full_rate = true;
 			break;
@@ -898,7 +887,7 @@ static int trigger_local_ho_or_as(struct ho_candidate *c, uint8_t requirements)
 		}
 		if (c->current.bts == c->target.bts) {
 			LOGPHOLCHAN(c->current.lchan, LOGL_INFO, "Not performing assignment: Already on target type\n");
-			return 0;
+			return -EALREADY;
 		}
 		break;
 	default:
@@ -907,23 +896,25 @@ static int trigger_local_ho_or_as(struct ho_candidate *c, uint8_t requirements)
 	}
 
 	/* trigger handover or assignment */
-	if  (c->current.bts == c->target.bts)
+	if  (c->current.bts == c->target.bts) {
 		LOGPHOLCHAN(c->current.lchan, LOGL_NOTICE, "Triggering assignment to %s, due to %s\n",
 			    full_rate ? "TCH/F" : "TCH/H",
 			    ho_reason_name(global_ho_reason));
-	else
+		return reassignment_request_to_chan_type(ASSIGN_FOR_CONGESTION_RESOLUTION, c->current.lchan,
+							 full_rate? GSM_LCHAN_TCH_F : GSM_LCHAN_TCH_H);
+	} else {
+		struct handover_out_req req = {
+			.from_hodec_id = HODEC2,
+			.old_lchan = c->current.lchan,
+			.new_lchan_type = full_rate? GSM_LCHAN_TCH_F : GSM_LCHAN_TCH_H,
+		};
+		bts_cell_ab(&req.target_cell_ab, c->target.bts);
 		LOGPHOLCHANTOBTS(c->current.lchan, c->target.bts, LOGL_INFO,
 				 "Triggering handover to %s, due to %s\n",
 				 full_rate ? "TCH/F" : "TCH/H",
 				 ho_reason_name(global_ho_reason));
-
-	req = (struct handover_out_req){
-		.from_hodec_id = HODEC2,
-		.old_lchan = c->current.lchan,
-		.target_nik = *bts_ident_key(c->target.bts),
-		.new_lchan_type = full_rate? GSM_LCHAN_TCH_F : GSM_LCHAN_TCH_H,
-	};
-	handover_request(&req);
+		handover_request(&req);
+	}
 	return 0;
 }
 
@@ -931,14 +922,14 @@ static int trigger_remote_bss_ho(struct ho_candidate *c, uint8_t requirements)
 {
 	struct handover_out_req req;
 
-	LOGPHOLCHANTOREMOTE(c->current.lchan, c->target.cil, LOGL_INFO,
+	LOGPHOLCHANTOREMOTE(c->current.lchan, &c->target.cell_ids, LOGL_INFO,
 			    "Triggering inter-BSC handover, due to %s\n",
 			    ho_reason_name(global_ho_reason));
 
 	req = (struct handover_out_req){
 		.from_hodec_id = HODEC2,
 		.old_lchan = c->current.lchan,
-		.target_nik = c->target.nik,
+		.target_cell_ab = c->target.ab,
 	};
 	handover_request(&req);
 	return 0;
@@ -974,13 +965,13 @@ static inline void debug_candidate(struct ho_candidate *candidate)
 	     candidate->target.free_tch##tchx, candidate->target.min_free_tch##tchx, \
 	     REQUIREMENTS_ARGS(candidate->requirements, TCHX)
 
-	if (!candidate->target.bts && !candidate->target.cil)
+	if (!candidate->target.bts && !candidate->target.cell_ids.id_list_len)
 		LOGPHOLCHAN(candidate->current.lchan, LOGL_DEBUG, "Empty candidate\n");
-	if (candidate->target.bts && candidate->target.cil)
+	if (candidate->target.bts && candidate->target.cell_ids.id_list_len)
 		LOGPHOLCHAN(candidate->current.lchan, LOGL_ERROR, "Invalid candidate: both local- and remote-BSS target\n");
 
-	if (candidate->target.cil)
-		LOGPHOLCHANTOREMOTE(candidate->current.lchan, candidate->target.cil, LOGL_DEBUG,
+	if (candidate->target.cell_ids.id_list_len)
+		LOGPHOLCHANTOREMOTE(candidate->current.lchan, &candidate->target.cell_ids, LOGL_DEBUG,
 				    "RX level %d dBm -> %d dBm\n",
 				    rxlev2dbm(candidate->current.rxlev), rxlev2dbm(candidate->target.rxlev));
 
@@ -1000,18 +991,21 @@ static inline void debug_candidate(struct ho_candidate *candidate)
 
 static void candidate_set_free_tch(struct ho_candidate *c)
 {
+	struct chan_counts bts_counts;
 	struct gsm_lchan *next_lchan;
 
-	c->current.free_tchf = bts_count_free_ts(c->current.bts, GSM_PCHAN_TCH_F);
+	chan_counts_for_bts(&bts_counts, c->current.bts);
+	c->current.free_tchf = bts_counts.val[CHAN_COUNTS1_ALL][CHAN_COUNTS2_FREE][GSM_LCHAN_TCH_F];
 	c->current.min_free_tchf = ho_get_hodec2_tchf_min_slots(c->current.bts->ho);
-	c->current.free_tchh = bts_count_free_ts(c->current.bts, GSM_PCHAN_TCH_H);
+	c->current.free_tchh = bts_counts.val[CHAN_COUNTS1_ALL][CHAN_COUNTS2_FREE][GSM_LCHAN_TCH_H];
 	c->current.min_free_tchh = ho_get_hodec2_tchh_min_slots(c->current.bts->ho);
+
 	switch (c->current.lchan->ts->pchan_is) {
 	case GSM_PCHAN_TCH_F:
 		c->current.free_tch = c->current.free_tchf;
 		c->current.min_free_tch = c->current.min_free_tchf;
 		c->current.lchan_frees_tchf = 1;
-		if (c->current.lchan->ts->pchan_on_init == GSM_PCHAN_TCH_F_TCH_H_PDCH)
+		if (c->current.lchan->ts->pchan_on_init == GSM_PCHAN_OSMO_DYN)
 			c->current.lchan_frees_tchh = 2;
 		else
 			c->current.lchan_frees_tchh = 0;
@@ -1023,7 +1017,7 @@ static void candidate_set_free_tch(struct ho_candidate *c)
 		/* Freeing one of two TCH/H does not free a dyn TS and would not free a TCH/F. It has to be the last
 		 * TCH/H of a dynamic timeslot that is freed to get a new TCH/F in the current cell from the handover.
 		 * Hence the ts_usage_count() condition. */
-		if (c->current.lchan->ts->pchan_on_init == GSM_PCHAN_TCH_F_TCH_H_PDCH
+		if (c->current.lchan->ts->pchan_on_init == GSM_PCHAN_OSMO_DYN
 		    && ts_usage_count(c->current.lchan->ts) == 1)
 			c->current.lchan_frees_tchf = 1;
 		else
@@ -1033,14 +1027,15 @@ static void candidate_set_free_tch(struct ho_candidate *c)
 		break;
 	}
 
-	c->target.free_tchf = bts_count_free_ts(c->target.bts, GSM_PCHAN_TCH_F);
+	chan_counts_for_bts(&bts_counts, c->target.bts);
+	c->target.free_tchf = bts_counts.val[CHAN_COUNTS1_ALL][CHAN_COUNTS2_FREE][GSM_LCHAN_TCH_F];
 	c->target.min_free_tchf = ho_get_hodec2_tchf_min_slots(c->target.bts->ho);
-	c->target.free_tchh = bts_count_free_ts(c->target.bts, GSM_PCHAN_TCH_H);
+	c->target.free_tchh = bts_counts.val[CHAN_COUNTS1_ALL][CHAN_COUNTS2_FREE][GSM_LCHAN_TCH_H];
 	c->target.min_free_tchh = ho_get_hodec2_tchh_min_slots(c->target.bts->ho);
 
 	/* Would the next TCH/F lchan occupy a dynamic timeslot that currently counts for free TCH/H timeslots? */
 	next_lchan = lchan_avail_by_type(c->target.bts, GSM_LCHAN_TCH_F, false);
-	if (next_lchan && next_lchan->ts->pchan_on_init == GSM_PCHAN_TCH_F_TCH_H_PDCH)
+	if (next_lchan && next_lchan->ts->pchan_on_init == GSM_PCHAN_OSMO_DYN)
 		c->target.next_tchf_reduces_tchh = 2;
 	else
 		c->target.next_tchf_reduces_tchh = 0;
@@ -1048,7 +1043,7 @@ static void candidate_set_free_tch(struct ho_candidate *c)
 	/* Would the next TCH/H lchan occupy a dynamic timeslot that currently counts for free TCH/F timeslots?
 	 * Note that a dyn TS already in TCH/H mode (half occupied) would not reduce free TCH/F. */
 	next_lchan = lchan_avail_by_type(c->target.bts, GSM_LCHAN_TCH_H, false);
-	if (next_lchan && next_lchan->ts->pchan_on_init == GSM_PCHAN_TCH_F_TCH_H_PDCH
+	if (next_lchan && next_lchan->ts->pchan_on_init == GSM_PCHAN_OSMO_DYN
 	    && next_lchan->ts->pchan_is != GSM_PCHAN_TCH_H)
 		c->target.next_tchh_reduces_tchf = 1;
 	else
@@ -1093,14 +1088,12 @@ static void collect_handover_candidate(struct gsm_lchan *lchan, struct neigh_mea
 {
 	struct gsm_bts *bts = lchan->ts->trx->bts;
 	struct gsm_bts *neighbor_bts;
-	const struct gsm0808_cell_id_list2 *neighbor_cil;
-	struct neighbor_ident_key ni = {
-		.from_bts = bts->nr,
+	struct gsm0808_cell_id_list2 neighbor_cil;
+	struct cell_ab target_ab = {
 		.arfcn = nmp->arfcn,
 		.bsic = nmp->bsic,
 	};
 	struct ho_candidate c;
-	int min_rxlev;
 	struct handover_cfg *neigh_cfg;
 
 	/* skip empty slots */
@@ -1119,9 +1112,9 @@ static void collect_handover_candidate(struct gsm_lchan *lchan, struct neigh_mea
 	}
 
 	find_handover_target_cell(&neighbor_bts, &neighbor_cil,
-				  lchan->conn, &ni, false);
+				  lchan->conn, &target_ab, false);
 
-	if (!neighbor_bts && !neighbor_cil) {
+	if (!neighbor_bts && !neighbor_cil.id_list_len) {
 		LOGPHOBTS(bts, LOGL_DEBUG, "no neighbor ARFCN %u BSIC %u configured for this cell\n",
 			  nmp->arfcn, nmp->bsic);
 		return;
@@ -1144,9 +1137,9 @@ static void collect_handover_candidate(struct gsm_lchan *lchan, struct neigh_mea
 			.rxlev = rxlev_current,
 		},
 		.target = {
-			.nik = ni,
+			.ab = target_ab,
 			.bts = neighbor_bts,
-			.cil = neighbor_cil,
+			.cell_ids = neighbor_cil,
 			.rxlev = neigh_meas_avg(nmp, ho_get_hodec2_rxlev_neigh_avg_win(bts->ho)),
 		},
 	};
@@ -1168,12 +1161,11 @@ static void collect_handover_candidate(struct gsm_lchan *lchan, struct neigh_mea
 
 	/* if the minimum level is not reached.
 	 * In case of a remote-BSS, use the current BTS' configuration. */
-	min_rxlev = ho_get_hodec2_min_rxlev(neigh_cfg);
-	if (rxlev2dbm(c.target.rxlev) < min_rxlev) {
+	if (is_low_rxlev(c.target.rxlev, neigh_cfg)) {
 		LOGPHOCAND(&c, LOGL_DEBUG,
 			   "Not a candidate, because RX level (%d dBm) is lower"
 			   " than the minimum required RX level (%d dBm)\n",
-			   rxlev2dbm(c.target.rxlev), min_rxlev);
+			   rxlev2dbm(c.target.rxlev), ho_get_hodec2_min_rxlev(neigh_cfg));
 		return;
 	}
 
@@ -1218,7 +1210,12 @@ static void collect_candidates_for_lchan(struct gsm_lchan *lchan,
 	assignment = ho_get_hodec2_as_active(bts->ho);
 	handover = ho_get_ho_active(bts->ho);
 
-	if (assignment)
+	/* See if re-assignment within the same cell can resolve congestion.
+	 * But: when TCH/F has low rxlev, do not re-assign. If a low rxlev TCH/F were re-assigned to TCH/H, we would
+	 * subsequently oscillate back to TCH/F due to low rxlev. So skip TCH/F with low rxlev. */
+	if (assignment
+	    && !(lchan->type == GSM_LCHAN_TCH_F
+		 && (is_low_rxlev(rxlev_current, bts->ho) || is_low_rxqual(current_rxqual(lchan), bts->ho))))
 		collect_assignment_candidate(lchan, clist, candidates, rxlev_current);
 
 	if (handover) {
@@ -1282,10 +1279,10 @@ static void collect_candidates_for_lchan(struct gsm_lchan *lchan,
  * If minimum RXLEV, minimum RXQUAL or maximum TA are exceeded, the caller should pass
  * include_weaker_rxlev=true so that handover is performed despite congestion.
  */
-static int find_alternative_lchan(struct gsm_lchan *lchan, bool include_weaker_rxlev)
+static int find_alternative_lchan(struct gsm_lchan *lchan, bool include_weaker_rxlev, bool request_upgrade_to_tch_f)
 {
 	struct gsm_bts *bts = lchan->ts->trx->bts;
-	int ahs = (lchan->tch_mode == GSM48_CMODE_SPEECH_AMR
+	int ahs = (gsm48_chan_mode_to_non_vamos(lchan->current_ch_mode_rate.chan_mode) == GSM48_CMODE_SPEECH_AMR
 		   && lchan->type == GSM_LCHAN_TCH_H);
 	int rxlev_current;
 	struct ho_candidate clist[1 + ARRAY_SIZE(lchan->neigh_meas)];
@@ -1326,9 +1323,10 @@ static int find_alternative_lchan(struct gsm_lchan *lchan, bool include_weaker_r
 			continue;
 
 		better = clist[i].target.rxlev - clist[i].current.rxlev;
-		/* Apply AFS bias? */
+		/* Apply AFS bias? Skip AFS bias for all intra-cell candidates. */
 		afs_bias = 0;
-		if (ahs && (clist[i].requirements & REQUIREMENT_B_TCHF))
+		if (clist[i].target.bts != bts
+		    && ahs && (clist[i].requirements & REQUIREMENT_B_TCHF))
 			afs_bias = ho_get_hodec2_afs_bias_rxlev(clist[i].target.bts->ho);
 		better += afs_bias;
 		if (better > best_better_db) {
@@ -1359,9 +1357,10 @@ static int find_alternative_lchan(struct gsm_lchan *lchan, bool include_weaker_r
 			continue;
 
 		better = clist[i].target.rxlev - clist[i].current.rxlev;
-		/* Apply AFS bias? */
+		/* Apply AFS bias? Skip AFS bias for all intra-cell candidates. */
 		afs_bias = 0;
-		if (ahs && (clist[i].requirements & REQUIREMENT_C_TCHF))
+		if (clist[i].target.bts != bts
+		    && ahs && (clist[i].requirements & REQUIREMENT_C_TCHF))
 			afs_bias = ho_get_hodec2_afs_bias_rxlev(clist[i].target.bts->ho);
 		better += afs_bias;
 		if (better > best_better_db) {
@@ -1396,14 +1395,18 @@ static int find_alternative_lchan(struct gsm_lchan *lchan, bool include_weaker_r
 			continue;
 
 		better = clist[i].target.rxlev - clist[i].current.rxlev;
-		/* Apply AFS bias?
+		/* Apply AFS bias? Skip AFS bias for all intra-cell candidates.
 		 * (never to remote-BSS neighbors, since we will not change the lchan type for those.) */
 		afs_bias = 0;
 		if (ahs && (clist[i].requirements & REQUIREMENT_A_TCHF)
-		    && clist[i].target.bts)
+		    && clist[i].target.bts && clist[i].target.bts != bts)
 			afs_bias = ho_get_hodec2_afs_bias_rxlev(clist[i].target.bts->ho);
 		better += afs_bias;
-		if (better > best_better_db) {
+		if (better > best_better_db
+		    || (better >= best_better_db /* Upgrade from TCH/H to TCH/F: allow for equal rxlev */
+			&& request_upgrade_to_tch_f
+			&& is_upgrade_to_tchf(&clist[i], REQUIREMENT_A_MASK))) {
+
 			best_cand = &clist[i];
 			best_better_db = better;
 			best_applied_afs_bias = afs_bias? true : false;
@@ -1412,10 +1415,21 @@ static int find_alternative_lchan(struct gsm_lchan *lchan, bool include_weaker_r
 
 	/* perform handover, if there is a candidate */
 	if (best_cand) {
+		int rc;
 		LOGPHOCAND(best_cand, LOGL_INFO, "Best candidate: RX level %d%s\n",
 			   rxlev2dbm(best_cand->target.rxlev),
 			   best_applied_afs_bias ? " (applied AHS -> AFS rxlev bias)" : "");
-		return trigger_ho(best_cand, best_cand->requirements & REQUIREMENT_A_MASK);
+		rc = trigger_ho(best_cand, best_cand->requirements & REQUIREMENT_A_MASK);
+
+		/* After upgrading TCH/H to TCH/F due to bad RxQual, start penalty timer to avoid re-assignment within
+		 * the same cell again, to avoid oscillation from RxQual noise combined with congestion resolution. */
+		if (!rc && is_upgrade_to_tchf(best_cand, REQUIREMENT_A_MASK)) {
+			struct gsm0808_cell_id bts_id;
+			gsm_bts_cell_id(&bts_id, best_cand->target.bts);
+			penalty_timers_add(lchan->conn, &lchan->conn->hodec2.penalty_timers, &bts_id,
+					   ho_get_hodec2_penalty_low_rxqual_as(bts->ho));
+		}
+		return rc;
 	}
 
 	/* Damn, all is congested, has too low RXLEV or cannot service the voice call due to codec
@@ -1481,14 +1495,8 @@ static void on_measurement_report(struct gsm_meas_rep *mr)
 	}
 
 	/* get average levels. if not enough measurements yet, value is < 0 */
-	av_rxlev = get_meas_rep_avg(lchan,
-				    ho_get_hodec2_full_tdma(bts->ho) ?
-				    MEAS_REP_DL_RXLEV_FULL : MEAS_REP_DL_RXLEV_SUB,
-				    ho_get_hodec2_rxlev_avg_win(bts->ho));
-	av_rxqual = get_meas_rep_avg(lchan,
-				     ho_get_hodec2_full_tdma(bts->ho) ?
-				     MEAS_REP_DL_RXQUAL_FULL : MEAS_REP_DL_RXQUAL_SUB,
-				     ho_get_hodec2_rxqual_avg_win(bts->ho));
+	av_rxlev = current_rxlev(lchan);
+	av_rxqual = current_rxqual(lchan);
 	if (av_rxlev < 0 && av_rxqual < 0) {
 		LOGPHOLCHAN(lchan, LOGL_INFO, "Skipping, Not enough recent measurements\n");
 		return;
@@ -1496,7 +1504,7 @@ static void on_measurement_report(struct gsm_meas_rep *mr)
 
 	/* improve levels in case of AFS, if defined */
 	if (lchan->type == GSM_LCHAN_TCH_F
-	 && lchan->tch_mode == GSM48_CMODE_SPEECH_AMR) {
+	 && gsm48_chan_mode_to_non_vamos(lchan->current_ch_mode_rate.chan_mode) == GSM48_CMODE_SPEECH_AMR) {
 		int av_rxlev_was = av_rxlev;
 		int av_rxqual_was = av_rxqual;
 		int rxlev_bias = ho_get_hodec2_afs_bias_rxlev(bts->ho);
@@ -1528,22 +1536,23 @@ static void on_measurement_report(struct gsm_meas_rep *mr)
 			global_ho_reason = HO_REASON_BAD_QUALITY;
 			LOGPHOLCHAN(lchan, LOGL_INFO, "Trying handover/assignment due to bad quality\n");
 		}
-		find_alternative_lchan(lchan, true);
+		find_alternative_lchan(lchan, true, true);
 		return;
 	}
 
 	/* Low Level */
-	if (av_rxlev >= 0 && rxlev2dbm(av_rxlev) < ho_get_hodec2_min_rxlev(bts->ho)) {
+	if (is_low_rxlev(av_rxlev, bts->ho)) {
 		global_ho_reason = HO_REASON_LOW_RXLEVEL;
 		LOGPHOLCHAN(lchan, LOGL_NOTICE, "RX level is TOO LOW: %d < %d\n",
 			    rxlev2dbm(av_rxlev), ho_get_hodec2_min_rxlev(bts->ho));
-		find_alternative_lchan(lchan, true);
+		find_alternative_lchan(lchan, true, true);
 		return;
 	}
 
 	/* Max Distance */
 	if (lchan->meas_rep_count > 0
 	    && lchan->last_ta > ho_get_hodec2_max_distance(bts->ho)) {
+		struct gsm0808_cell_id bts_id;
 		global_ho_reason = HO_REASON_MAX_DISTANCE;
 		LOGPHOLCHAN(lchan, LOGL_NOTICE, "TA is TOO HIGH: %u > %d\n",
 			    lchan->last_ta, ho_get_hodec2_max_distance(bts->ho));
@@ -1551,8 +1560,10 @@ static void on_measurement_report(struct gsm_meas_rep *mr)
 		 * early. it must be started before selecting a better cell,
 		 * so there is no assignment selected, due to running
 		 * penalty timer. */
-		bts_penalty_time_add(lchan->conn, bts, ho_get_hodec2_penalty_max_dist(bts->ho));
-		find_alternative_lchan(lchan, true);
+		gsm_bts_cell_id(&bts_id, bts);
+		penalty_timers_add(lchan->conn, &lchan->conn->hodec2.penalty_timers, &bts_id,
+				   ho_get_hodec2_penalty_max_dist(bts->ho));
+		find_alternative_lchan(lchan, true, false);
 		return;
 	}
 
@@ -1563,23 +1574,16 @@ static void on_measurement_report(struct gsm_meas_rep *mr)
 	/* try handover to a better cell */
 	if (av_rxlev >= 0 && (mr->nr % pwr_interval) == 0) {
 		global_ho_reason = HO_REASON_BETTER_CELL;
-		find_alternative_lchan(lchan, false);
+		find_alternative_lchan(lchan, false, false);
 	}
 }
 
 static bool lchan_is_on_dynamic_ts(struct gsm_lchan *lchan)
 {
-	return lchan->ts->pchan_on_init == GSM_PCHAN_TCH_F_TCH_H_PDCH
+	return lchan->ts->pchan_on_init == GSM_PCHAN_OSMO_DYN
 		|| lchan->ts->pchan_on_init == GSM_PCHAN_TCH_F_PDCH;
 }
 
-static bool is_upgrade_to_tchf(const struct ho_candidate *c, uint8_t for_requirement)
-{
-	return c->current.lchan
-		&& (c->current.lchan->type == GSM_LCHAN_TCH_H)
-		&& ((c->requirements & for_requirement) & (REQUIREMENT_B_TCHF | REQUIREMENT_C_TCHF));
-}
-
 /* Given two candidates, pick the one that should rather be moved during handover.
  * Return the better candidate in out-parameters best_cand and best_avg_db.
  */
@@ -1929,6 +1933,7 @@ exit:
 
 static void bts_congestion_check(struct gsm_bts *bts)
 {
+	struct chan_counts bts_counts;
 	int min_free_tchf, min_free_tchh;
 	int free_tchf, free_tchh;
 
@@ -1956,8 +1961,9 @@ static void bts_congestion_check(struct gsm_bts *bts)
 		return;
 	}
 
-	free_tchf = bts_count_free_ts(bts, GSM_PCHAN_TCH_F);
-	free_tchh = bts_count_free_ts(bts, GSM_PCHAN_TCH_H);
+	chan_counts_for_bts(&bts_counts, bts);
+	free_tchf = bts_counts.val[CHAN_COUNTS1_ALL][CHAN_COUNTS2_FREE][GSM_LCHAN_TCH_F];
+	free_tchh = bts_counts.val[CHAN_COUNTS1_ALL][CHAN_COUNTS2_FREE][GSM_LCHAN_TCH_H];
 	LOGPHOBTS(bts, LOGL_INFO, "Congestion check: (free/want-free) TCH/F=%d/%d TCH/H=%d/%d\n",
 		  free_tchf, min_free_tchf, free_tchh, min_free_tchh);
 
@@ -1989,7 +1995,6 @@ static void congestion_check_cb(void *arg)
 static void on_handover_end(struct gsm_subscriber_connection *conn, enum handover_result result)
 {
 	struct gsm_bts *old_bts = NULL;
-	struct gsm_bts *new_bts = NULL;
 	int penalty;
 	struct handover *ho = &conn->ho;
 
@@ -1999,8 +2004,6 @@ static void on_handover_end(struct gsm_subscriber_connection *conn, enum handove
 
 	if (conn->lchan)
 		old_bts = conn->lchan->ts->trx->bts;
-	if (ho->new_lchan)
-		new_bts = ho->new_lchan->ts->trx->bts;
 
 	/* Only interested in handovers within this BSS or going out into another BSS. Incoming handovers
 	 * from another BSS are accounted for in the other BSS. */
@@ -2027,11 +2030,7 @@ static void on_handover_end(struct gsm_subscriber_connection *conn, enum handove
 
 	LOG_HO(conn, LOGL_NOTICE, "Failed, starting penalty timer (%d s)\n", penalty);
 	conn->hodec2.failures = 0;
-
-	if (new_bts)
-		bts_penalty_time_add(conn, new_bts, penalty);
-	else
-		nik_penalty_time_add(conn, &ho->target_cell, penalty);
+	penalty_timers_add_list(conn, &conn->hodec2.penalty_timers, &ho->target_cell_ids, penalty);
 }
 
 static struct handover_decision_callbacks hodec2_callbacks = {
