@@ -34,6 +34,7 @@
 #include <osmocom/bsc/abis_rsl.h>
 #include <osmocom/bsc/handover_fsm.h>
 #include <osmocom/bsc/bts.h>
+#include <osmocom/bsc/bsc_stats.h>
 
 #include <osmocom/gsm/protocol/gsm_08_08.h>
 #include <osmocom/gsm/gsm0808.h>
@@ -343,7 +344,7 @@ static int bssmap_handle_paging(struct bsc_msc_data *msc,
 
 int bsc_paging_start(struct bsc_paging_params *params)
 {
-	rate_ctr_inc(&bsc_gsmnet->bsc_ctrs->ctr[BSC_CTR_PAGING_ATTEMPTED]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(bsc_gsmnet->bsc_ctrs, BSC_CTR_PAGING_ATTEMPTED));
 
 	if (!params->bsub) {
 		params->bsub = bsc_subscr_find_or_create_by_imsi(bsc_gsmnet->bsc_subscribers, params->imsi.imsi,
@@ -416,55 +417,31 @@ static int select_best_cipher(uint8_t msc_mask, uint8_t bsc_mask)
 	return -1;
 }
 
-/*! We received a GSM 08.08 CIPHER MODE from the MSC */
-static int gsm0808_cipher_mode(struct gsm_subscriber_connection *conn, int cipher,
-			       const uint8_t *key, int len, int include_imeisv)
-{
-	if (cipher > 0 && key == NULL) {
-		LOGP(DRSL, LOGL_ERROR, "%s: Need to have an encryption key.\n",
-		     bsc_subscr_name(conn->bsub));
-		return -1;
-	}
-
-	if (len > MAX_A5_KEY_LEN) {
-		LOGP(DRSL, LOGL_ERROR, "%s: The key is too long: %d\n",
-		     bsc_subscr_name(conn->bsub), len);
-		return -1;
-	}
-
-	LOGP(DRSL, LOGL_DEBUG, "(subscr %s) Cipher Mode: cipher=%d key=%s include_imeisv=%d\n",
-	     bsc_subscr_name(conn->bsub), cipher, osmo_hexdump_nospc(key, len), include_imeisv);
-
-	conn->lchan->encr.alg_id = RSL_ENC_ALG_A5(cipher);
-	if (key) {
-		conn->lchan->encr.key_len = len;
-		memcpy(conn->lchan->encr.key, key, len);
-	}
-
-	return gsm48_send_rr_ciph_mode(conn->lchan, include_imeisv);
-}
-
 static int bssmap_handle_clear_cmd(struct gsm_subscriber_connection *conn,
 				   struct msgb *msg, unsigned int length)
 {
 	struct tlv_parsed tp;
-	struct gscon_clear_cmd_data ccd = {
-		.is_csfb = false,
-	};
+	enum gsm0808_cause cause_0808;
 
 	tlv_parse(&tp, gsm0808_att_tlvdef(), msg->l4h + 1, length - 1, 0, 0);
 
-	ccd.cause_0808 = gsm0808_get_cause(&tp);
-	if (ccd.cause_0808 < 0) {
+	cause_0808 = gsm0808_get_cause(&tp);
+	if (cause_0808 < 0) {
 		LOGPFSML(conn->fi, LOGL_ERROR, "Clear Command: Mandatory Cause IE not present.\n");
 		/* Clear anyway, but without a proper cause. */
-		ccd.cause_0808 = GSM0808_CAUSE_RADIO_INTERFACE_MESSAGE_FAILURE;
+		cause_0808 = GSM0808_CAUSE_RADIO_INTERFACE_MESSAGE_FAILURE;
 	}
 
-	if (TLVP_PRESENT(&tp, GSM0808_IE_CSFB_INDICATION))
-		ccd.is_csfb = true;
+	if (TLVP_PRESENT(&tp, GSM0808_IE_CSFB_INDICATION) &&
+	    !conn->fast_return.last_eutran_plmn_valid) {
+		LOGPFSML(conn->fi, LOGL_NOTICE,
+			 "Clear Command: CSFB Indication present, "
+			 "but subscriber has no Last Used E-UTRAN PLMN Id! "
+			 "This probably means MSC doesn't support proper return "
+			 "to the last used PLMN after CS fallback.\n");
+	}
 
-	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_A_CLEAR_CMD, &ccd);
+	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_A_CLEAR_CMD, &cause_0808);
 
 	return 0;
 }
@@ -490,6 +467,7 @@ static int bssmap_handle_cipher_mode(struct gsm_subscriber_connection *conn,
 	uint16_t enc_key_len;
 	uint8_t enc_bits_msc;
 	int chosen_cipher;
+	const struct tlv_p_entry *ie_kc128;
 
 	if (!conn || !conn->lchan) {
 		LOGP(DMSC, LOGL_ERROR, "No lchan/msc_data in cipher mode command.\n");
@@ -549,13 +527,51 @@ static int bssmap_handle_cipher_mode(struct gsm_subscriber_connection *conn,
 		goto reject;
 	}
 
-	/* To complete the confusion, gsm0808_cipher_mode again expects the encryption as a number
-	 * from 0 to 7. */
-	if (gsm0808_cipher_mode(conn, chosen_cipher, enc_key, enc_key_len,
-				include_imeisv)) {
+	if (chosen_cipher > 0 && !enc_key_len) {
+		LOGP(DRSL, LOGL_ERROR, "%s: Need to have an encryption key.\n",
+		     bsc_subscr_name(conn->bsub));
 		reject_cause = GSM0808_CAUSE_PROTOCOL_ERROR_BETWEEN_BSS_AND_MSC;
 		goto reject;
 	}
+
+	if (enc_key_len > MAX_A5_KEY_LEN) {
+		LOGP(DRSL, LOGL_ERROR, "%s: The key is too long: %d\n",
+		     bsc_subscr_name(conn->bsub), len);
+		reject_cause = GSM0808_CAUSE_PROTOCOL_ERROR_BETWEEN_BSS_AND_MSC;
+		goto reject;
+	}
+
+	conn->lchan->encr.alg_id = RSL_ENC_ALG_A5(chosen_cipher);
+	if (enc_key_len) {
+		conn->lchan->encr.key_len = enc_key_len;
+		memcpy(conn->lchan->encr.key, enc_key, enc_key_len);
+	}
+	if ((ie_kc128 = TLVP_GET(&tp, GSM0808_IE_KC_128))) {
+		if (ie_kc128->len != sizeof(conn->lchan->encr.kc128)) {
+			LOGPFSML(conn->fi, LOGL_ERROR, "Kc128 IE has wrong length: %u (expect %zu)\n",
+				 ie_kc128->len, sizeof(conn->lchan->encr.kc128));
+			reject_cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
+			goto reject;
+		}
+		memcpy(conn->lchan->encr.kc128, ie_kc128->val, sizeof(conn->lchan->encr.kc128));
+		conn->lchan->encr.kc128_present = true;
+	}
+
+	if (chosen_cipher == 4 && !conn->lchan->encr.kc128_present) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "A5/4 encryption selected, but no Kc128\n");
+		reject_cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
+		goto reject;
+	}
+
+	LOGP(DRSL, LOGL_DEBUG, "(subscr %s) Cipher Mode: cipher=%d key=%s kc128=%s include_imeisv=%d\n",
+	     bsc_subscr_name(conn->bsub), chosen_cipher, osmo_hexdump_nospc(enc_key, enc_key_len),
+	     ie_kc128? osmo_hexdump_nospc_c(OTC_SELECT, ie_kc128->val, ie_kc128->len) : "-",
+	     include_imeisv);
+
+	if (gsm48_send_rr_ciph_mode(conn->lchan, include_imeisv) < 0) {
+		reject_cause = GSM0808_CAUSE_RADIO_INTERFACE_FAILURE;
+		goto reject;
+	}
 	return 0;
 
 reject:
@@ -565,7 +581,7 @@ reject:
 		return -1;
 	}
 
-	rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_CIPHER_REJECT]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_CIPHER_REJECT));
 	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
 	return -1;
 }
@@ -646,7 +662,7 @@ static int bssmap_handle_lcls_connect_ctrl(struct gsm_subscriber_connection *con
 	LOGPFSM(conn->fi, "Tx LCLS CONNECT CTRL ACK (%s)\n",
 		gsm0808_lcls_status_name(lcls_get_status(conn)));
 	resp = gsm0808_create_lcls_conn_ctrl_ack(lcls_get_status(conn));
-	rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_LCLS_CONNECT_CTRL_ACK]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_LCLS_CONNECT_CTRL_ACK));
 	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
 
 	return 0;
@@ -675,34 +691,34 @@ static int select_codecs(struct assignment_request *req, struct gsm0808_channel_
 
 	switch (ct->ch_rate_type) {
 	case GSM0808_SPEECH_FULL_BM:
-		rc = match_codec_pref(&req->ch_mode_rate[nc], ct, &conn->codec_list, msc, bts,
+		rc = match_codec_pref(&req->ch_mode_rate_list[nc], ct, &conn->codec_list, msc, bts,
 				      RATE_PREF_FR);
-		nc += (rc == 0);
+		nc += (rc == 0) ? 1 : 0;
 		break;
 	case GSM0808_SPEECH_HALF_LM:
-		rc = match_codec_pref(&req->ch_mode_rate[nc], ct, &conn->codec_list, msc, bts,
+		rc = match_codec_pref(&req->ch_mode_rate_list[nc], ct, &conn->codec_list, msc, bts,
 				      RATE_PREF_HR);
-		nc += (rc == 0);
+		nc += (rc == 0) ? 1 : 0;
 		break;
 	case GSM0808_SPEECH_PERM:
 	case GSM0808_SPEECH_PERM_NO_CHANGE:
 	case GSM0808_SPEECH_FULL_PREF_NO_CHANGE:
 	case GSM0808_SPEECH_FULL_PREF:
-		rc = match_codec_pref(&req->ch_mode_rate[nc], ct, &conn->codec_list, msc, bts,
+		rc = match_codec_pref(&req->ch_mode_rate_list[nc], ct, &conn->codec_list, msc, bts,
 				      RATE_PREF_FR);
-		nc += (rc == 0);
-		rc = match_codec_pref(&req->ch_mode_rate[nc], ct, &conn->codec_list, msc, bts,
+		nc += (rc == 0) ? 1 : 0;
+		rc = match_codec_pref(&req->ch_mode_rate_list[nc], ct, &conn->codec_list, msc, bts,
 				      RATE_PREF_HR);
-		nc += (rc == 0);
+		nc += (rc == 0) ? 1 : 0;
 		break;
 	case GSM0808_SPEECH_HALF_PREF_NO_CHANGE:
 	case GSM0808_SPEECH_HALF_PREF:
-		rc = match_codec_pref(&req->ch_mode_rate[nc], ct, &conn->codec_list, msc, bts,
+		rc = match_codec_pref(&req->ch_mode_rate_list[nc], ct, &conn->codec_list, msc, bts,
 				      RATE_PREF_HR);
-		nc += (rc == 0);
-		rc = match_codec_pref(&req->ch_mode_rate[nc], ct, &conn->codec_list, msc, bts,
+		nc += (rc == 0) ? 1 : 0;
+		rc = match_codec_pref(&req->ch_mode_rate_list[nc], ct, &conn->codec_list, msc, bts,
 				      RATE_PREF_FR);
-		nc += (rc == 0);
+		nc += (rc == 0) ? 1 : 0;
 		break;
 	default:
 		rc = -EINVAL;
@@ -722,8 +738,8 @@ static int select_codecs(struct assignment_request *req, struct gsm0808_channel_
 		DEBUGP(DMSC, "Found matching audio type (pref=%d): %s %s for channel_type ="
 		       " { ch_indctr=0x%x, ch_rate_type=0x%x, perm_spch=[ %s] }\n",
 		       i,
-		       req->ch_mode_rate[i].chan_rate == CH_RATE_FULL ? "full rate" : "half rate",
-		       get_value_string(gsm48_chan_mode_names, req->ch_mode_rate[i].chan_mode),
+		       req->ch_mode_rate_list[i].chan_rate == CH_RATE_FULL ? "full rate" : "half rate",
+		       get_value_string(gsm48_chan_mode_names, req->ch_mode_rate_list[i].chan_mode),
 		       ct->ch_indctr, ct->ch_rate_type, osmo_hexdump(ct->perm_spch, ct->perm_spch_len));
 	}
 
@@ -738,43 +754,43 @@ static int select_sign_chan(struct assignment_request *req, struct gsm0808_chann
 
 	switch (ct->ch_rate_type) {
 	case GSM0808_SIGN_ANY:
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_SDCCH;
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_HALF;
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_FULL;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_SDCCH;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_HALF;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_FULL;
 		break;
 	case GSM0808_SIGN_SDCCH:
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_SDCCH;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_SDCCH;
 		break;
 	case GSM0808_SIGN_SDCCH_FULL_BM:
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_SDCCH;
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_FULL;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_SDCCH;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_FULL;
 		break;
 	case GSM0808_SIGN_SDCCH_HALF_LM:
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_SDCCH;
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_HALF;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_SDCCH;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_HALF;
 		break;
 	case GSM0808_SIGN_FULL_BM:
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_FULL;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_FULL;
 		break;
 	case GSM0808_SIGN_HALF_LM:
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_HALF;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_HALF;
 		break;
 	case GSM0808_SIGN_FULL_PREF:
 	case GSM0808_SIGN_FULL_PREF_NO_CHANGE:
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_FULL;
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_HALF;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_FULL;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_HALF;
 		break;
 	case GSM0808_SIGN_HALF_PREF:
 	case GSM0808_SIGN_HALF_PREF_NO_CHANGE:
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_HALF;
-		req->ch_mode_rate[nc++].chan_rate = CH_RATE_FULL;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_HALF;
+		req->ch_mode_rate_list[nc++].chan_rate = CH_RATE_FULL;
 		break;
 	default:
 		break;
 	}
 
 	for (i = 0; i < nc; i++)
-		req->ch_mode_rate[i].chan_mode = GSM48_CMODE_SIGN;
+		req->ch_mode_rate_list[i].chan_mode = GSM48_CMODE_SIGN;
 
 	req->n_ch_mode_rate = nc;
 
@@ -919,6 +935,7 @@ static int bssmap_handle_assignm_req(struct gsm_subscriber_connection *conn,
 		}
 
 		req = (struct assignment_request){
+			.assign_for = ASSIGN_FOR_BSSMAP_REQ,
 			.aoip = aoip,
 			.msc_assigned_cic = cic,
 			.use_osmux = use_osmux,
@@ -947,6 +964,7 @@ static int bssmap_handle_assignm_req(struct gsm_subscriber_connection *conn,
 		break;
 	case GSM0808_CHAN_SIGN:
 		req = (struct assignment_request){
+			.assign_for = ASSIGN_FOR_BSSMAP_REQ,
 			.aoip = aoip,
 		};
 
@@ -967,7 +985,7 @@ reject:
 	resp = gsm0808_create_assignment_failure(cause, NULL);
 	OSMO_ASSERT(resp);
 
-	rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_ASSIGMENT_FAILURE]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_ASSIGMENT_FAILURE));
 	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
 	return -1;
 }
@@ -1072,7 +1090,6 @@ static int bssmap_handle_common_id(struct gsm_subscriber_connection *conn,
 				   struct msgb *msg, unsigned int length)
 {
 	struct tlv_parsed tp;
-	struct osmo_mobile_identity mi_imsi;
 
 	osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1);
 
@@ -1084,13 +1101,7 @@ static int bssmap_handle_common_id(struct gsm_subscriber_connection *conn,
 		return -EINVAL;
 	}
 
-	if (osmo_mobile_identity_decode(&mi_imsi, TLVP_VAL(&tp, GSM0808_IE_IMSI), TLVP_LEN(&tp, GSM0808_IE_IMSI), false)
-	    || mi_imsi.type != GSM_MI_TYPE_IMSI) {
-		LOGPFSML(conn->fi, LOGL_ERROR, "CommonID: could not parse IMSI\n");
-		return -EINVAL;
-	}
-
-	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_A_COMMON_ID_IND, &mi_imsi);
+	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_A_COMMON_ID_IND, &tp);
 
 	return 0;
 }
@@ -1361,6 +1372,7 @@ int bsc_tx_bssmap_ho_required(struct gsm_lchan *lchan, const struct gsm0808_cell
 {
 	int rc;
 	struct msgb *msg;
+	struct gsm_subscriber_connection *conn = lchan->conn;
 	struct gsm0808_handover_required params = {
 		.cause = GSM0808_CAUSE_BETTER_CELL,
 		.cil = *target_cells,
@@ -1368,12 +1380,21 @@ int bsc_tx_bssmap_ho_required(struct gsm_lchan *lchan, const struct gsm0808_cell
 		.current_channel_type_1 = gsm0808_current_channel_type_1(lchan->type),
 	};
 
+	/* Even if fast_return is now allowed locally, we may still want to
+	 * signal the Last EUTRAN PLMN Id to the new cell, since destination
+	 * config may differ and allow fast return */
+	if (conn->fast_return.last_eutran_plmn_valid) {
+		params.old_bss_to_new_bss_info_present = true;
+		params.old_bss_to_new_bss_info.last_eutran_plmn_id_present = true;
+		params.old_bss_to_new_bss_info.last_eutran_plmn_id = conn->fast_return.last_eutran_plmn;
+	}
+
 	switch (lchan->type) {
 	case GSM_LCHAN_TCH_F:
 	case GSM_LCHAN_TCH_H:
 		params.speech_version_used_present = true;
 		params.speech_version_used = gsm0808_permitted_speech(lchan->type,
-								      lchan->tch_mode);
+								      lchan->current_ch_mode_rate.chan_mode);
 		if (!params.speech_version_used) {
 			LOG_HO(lchan->conn, LOGL_ERROR, "Cannot encode Speech Version (Used)"
 			       " for BSSMAP Handover Required message\n");
@@ -1386,14 +1407,14 @@ int bsc_tx_bssmap_ho_required(struct gsm_lchan *lchan, const struct gsm0808_cell
 
 	msg = gsm0808_create_handover_required(&params);
 	if (!msg) {
-		LOG_HO(lchan->conn, LOGL_ERROR, "Cannot compose BSSMAP Handover Required message\n");
+		LOG_HO(conn, LOGL_ERROR, "Cannot compose BSSMAP Handover Required message\n");
 		return -EINVAL;
 	}
 
-	rate_ctr_inc(&lchan->conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_HANDOVER_REQUIRED]);
-	rc = gscon_sigtran_send(lchan->conn, msg);
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_HANDOVER_REQUIRED));
+	rc = gscon_sigtran_send(conn, msg);
 	if (rc) {
-		LOG_HO(lchan->conn, LOGL_ERROR, "Cannot send BSSMAP Handover Required message\n");
+		LOG_HO(conn, LOGL_ERROR, "Cannot send BSSMAP Handover Required message\n");
 		return rc;
 	}
 
@@ -1411,9 +1432,10 @@ int bsc_tx_bssmap_ho_request_ack(struct gsm_subscriber_connection *conn, struct
 		.l3_info = rr_ho_command->data,
 		.l3_info_len = rr_ho_command->len,
 		.chosen_channel_present = true,
-		.chosen_channel = gsm0808_chosen_channel(new_lchan->type, new_lchan->tch_mode),
+		.chosen_channel = gsm0808_chosen_channel(new_lchan->type, new_lchan->current_ch_mode_rate.chan_mode),
 		.chosen_encr_alg = new_lchan->encr.alg_id,
-		.chosen_speech_version = gsm0808_permitted_speech(new_lchan->type, new_lchan->tch_mode),
+		.chosen_speech_version = gsm0808_permitted_speech(new_lchan->type,
+								  new_lchan->current_ch_mode_rate.chan_mode),
 	};
 
 	if (gscon_is_aoip(conn)) {
@@ -1436,7 +1458,7 @@ int bsc_tx_bssmap_ho_request_ack(struct gsm_subscriber_connection *conn, struct
 		params.aoip_transport_layer = &ss;
 	}
 
-	rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_HANDOVER_RQST_ACKNOWLEDGE]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_HANDOVER_RQST_ACKNOWLEDGE));
 	LOG_HO(conn, LOGL_DEBUG, "Sending BSSMAP Handover Request Acknowledge\n");
 	msg = gsm0808_create_handover_request_ack2(&params);
 	msgb_free(rr_ho_command);
@@ -1452,7 +1474,7 @@ int bsc_tx_bssmap_ho_detect(struct gsm_subscriber_connection *conn)
 	if (!msg)
 		return -ENOMEM;
 
-	rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_HANDOVER_DETECT]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_HANDOVER_DETECT));
 	return osmo_bsc_sigtran_send(conn, msg);
 }
 
@@ -1469,7 +1491,7 @@ enum handover_result bsc_tx_bssmap_ho_complete(struct gsm_subscriber_connection
 		.chosen_encr_alg = lchan->encr.alg_id,
 
 		.chosen_channel_present = true,
-		.chosen_channel = gsm0808_chosen_channel(lchan->type, lchan->tch_mode),
+		.chosen_channel = gsm0808_chosen_channel(lchan->type, lchan->current_ch_mode_rate.chan_mode),
 
 		.lcls_bss_status_present = (lcls_status != 0xff),
 		.lcls_bss_status = lcls_status,
@@ -1477,7 +1499,7 @@ enum handover_result bsc_tx_bssmap_ho_complete(struct gsm_subscriber_connection
 
 	/* speech_codec_chosen */
 	if (ho->new_lchan->activate.info.requires_voice_stream && gscon_is_aoip(conn)) {
-		int perm_spch = gsm0808_permitted_speech(lchan->type, lchan->tch_mode);
+		int perm_spch = gsm0808_permitted_speech(lchan->type, lchan->current_ch_mode_rate.chan_mode);
 		params.speech_codec_chosen_present = true;
 		rc = gsm0808_speech_codec_from_chan_type(&params.speech_codec_chosen, perm_spch);
 		if (rc) {
@@ -1492,7 +1514,7 @@ enum handover_result bsc_tx_bssmap_ho_complete(struct gsm_subscriber_connection
 		return HO_RESULT_ERROR;
 	}
 
-	rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_HANDOVER_COMPLETE]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_HANDOVER_COMPLETE));
 	rc = osmo_bsc_sigtran_send(conn, msg);
 	if (rc) {
 		LOG_HO(conn, LOGL_ERROR, "Cannot send BSSMAP Handover Complete message\n");
@@ -1514,7 +1536,7 @@ void bsc_tx_bssmap_ho_failure(struct gsm_subscriber_connection *conn)
 		return;
 	}
 
-	rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_HANDOVER_FAILURE]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_HANDOVER_FAILURE));
 	rc = osmo_bsc_sigtran_send(conn, msg);
 	if (rc)
 		LOG_HO(conn, LOGL_ERROR, "Cannot send BSSMAP Handover Failure message (rc=%d %s)\n",
