@@ -52,7 +52,7 @@ static void ts_fsm_update_id(struct gsm_bts_trx_ts *ts)
 				  gsm_pchan_id(ts->pchan_on_init));
 }
 
-void ts_fsm_init()
+static __attribute__((constructor)) void ts_fsm_init(void)
 {
 	OSMO_ASSERT(osmo_fsm_register(&ts_fsm) == 0);
 }
@@ -66,6 +66,14 @@ void ts_fsm_alloc(struct gsm_bts_trx_ts *ts)
 	ts_fsm_update_id(ts);
 }
 
+void ts_fsm_free(struct gsm_bts_trx_ts *ts)
+{
+	if (ts->fi) {
+		osmo_fsm_inst_free(ts->fi);
+		ts->fi = NULL;
+	}
+}
+
 enum lchan_sanity {
 	LCHAN_IS_INSANE = -1,
 	LCHAN_IS_READY_TO_GO,
@@ -85,7 +93,9 @@ static enum lchan_sanity is_lchan_sane(struct gsm_bts_trx_ts *ts, struct gsm_lch
 		return LCHAN_IS_READY_TO_GO;
 
 	switch (ts->pchan_on_init) {
-	case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+	case GSM_PCHAN_OSMO_DYN:
+		if (lchan->type == GSM_LCHAN_SDCCH)
+			return LCHAN_NEEDS_PCHAN_CHANGE;
 		if (lchan->type == GSM_LCHAN_TCH_H)
 			return LCHAN_NEEDS_PCHAN_CHANGE;
 		/* fall thru */
@@ -112,7 +122,7 @@ static int ts_count_active_lchans(struct gsm_bts_trx_ts *ts)
 	struct gsm_lchan *lchan;
 	int count = 0;
 
-	ts_for_each_lchan(lchan, ts) {
+	ts_for_n_lchans(lchan, ts, ts->max_lchans_possible) {
 		if (lchan->fi->state == LCHAN_ST_UNUSED)
 			continue;
 		count++;
@@ -125,7 +135,7 @@ static void ts_lchans_dispatch(struct gsm_bts_trx_ts *ts, int lchan_state, uint3
 {
 	struct gsm_lchan *lchan;
 
-	ts_for_each_potential_lchan(lchan, ts) {
+	ts_for_n_lchans(lchan, ts, ts->max_lchans_possible) {
 		if (lchan_state >= 0
 		    && !lchan_state_is(lchan, lchan_state))
 			continue;
@@ -137,7 +147,7 @@ static void ts_terminate_lchan_fsms(struct gsm_bts_trx_ts *ts)
 {
 	struct gsm_lchan *lchan;
 
-	ts_for_each_potential_lchan(lchan, ts) {
+	ts_for_n_lchans(lchan, ts, ts->max_lchans_possible) {
 		osmo_fsm_inst_term(lchan->fi, OSMO_FSM_TERM_REQUEST, NULL);
 	}
 }
@@ -146,7 +156,7 @@ static int ts_lchans_waiting(struct gsm_bts_trx_ts *ts)
 {
 	struct gsm_lchan *lchan;
 	int count = 0;
-	ts_for_each_potential_lchan(lchan, ts)
+	ts_for_n_lchans(lchan, ts, ts->max_lchans_possible)
 		if (lchan->fi->state == LCHAN_ST_WAIT_TS_READY)
 			count++;
 	return count;
@@ -187,17 +197,59 @@ static void ts_fsm_err_ready_to_go_in_pdch(struct osmo_fsm_inst *fi, struct gsm_
 		     osmo_fsm_inst_state_name(fi), gsm_lchan_name(lchan));
 }
 
+void ts_set_pchan_is(struct gsm_bts_trx_ts *ts, enum gsm_phys_chan_config pchan_is)
+{
+	int i;
+	struct gsm_lchan *lchan;
+	uint8_t max_lchans_possible_vamos;
+
+	ts->pchan_is = pchan_is;
+	ts->max_primary_lchans = pchan_subslots(ts->pchan_is);
+	max_lchans_possible_vamos = pchan_subslots_vamos(ts->pchan_is);
+	LOG_TS(ts, LOGL_DEBUG, "pchan_is=%s max_primary_lchans=%d max_lchans_possible=%d (%u VAMOS)\n",
+	       gsm_pchan_name(ts->pchan_is), ts->max_primary_lchans, ts->max_lchans_possible,
+	       max_lchans_possible_vamos);
+	switch (ts->pchan_is) {
+	case GSM_PCHAN_TCH_F:
+	case GSM_PCHAN_TCH_H:
+		for (i = 0; i < ts->max_lchans_possible; i++) {
+			lchan = &ts->lchan[i];
+			if (i >= ts->max_primary_lchans &&
+			    (i - ts->max_primary_lchans) < (int)max_lchans_possible_vamos)
+				lchan->vamos.is_secondary = true;
+			else
+				lchan->vamos.is_secondary = false;
+			lchan_fsm_update_id(lchan);
+		}
+		break;
+	default:
+		ts_for_n_lchans(lchan, ts, ts->max_lchans_possible) {
+			lchan->vamos.is_secondary = false;
+			lchan_fsm_update_id(lchan);
+		}
+		break;
+	}
+}
+
 static void ts_setup_lchans(struct gsm_bts_trx_ts *ts)
 {
 	int i, max_lchans;
+	int max_lchans_vamos;
 
 	ts->pchan_on_init = ts->pchan_from_config;
 	ts_fsm_update_id(ts);
 
 	max_lchans = pchan_subslots(ts->pchan_on_init);
-	LOG_TS(ts, LOGL_DEBUG, "max lchans: %d\n", max_lchans);
+	if (osmo_bts_has_feature(&ts->trx->bts->features, BTS_FEAT_VAMOS))
+		max_lchans_vamos = pchan_subslots_vamos(ts->pchan_on_init);
+	else
+		max_lchans_vamos = 0;
+	LOG_TS(ts, LOGL_DEBUG, "max lchans: %d + %d VAMOS secondaries\n", max_lchans, max_lchans_vamos);
+	ts->max_lchans_possible = max_lchans + max_lchans_vamos;
+	ts->max_primary_lchans = 0;
 
-	for (i = 0; i < max_lchans; i++) {
+	OSMO_ASSERT(ts->max_lchans_possible <= TS_MAX_LCHAN);
+	for (i = 0; i < ts->max_lchans_possible; i++) {
 		/* If we receive more than one Channel OPSTART ACK, don't fail on the second init. */
 		if (ts->lchan[i].fi)
 			continue;
@@ -205,18 +257,16 @@ static void ts_setup_lchans(struct gsm_bts_trx_ts *ts)
 	}
 
 	switch (ts->pchan_on_init) {
-	case GSM_PCHAN_TCH_F_TCH_H_PDCH:
-		ts->pchan_is = GSM_PCHAN_NONE;
+	case GSM_PCHAN_OSMO_DYN:
+		ts_set_pchan_is(ts, GSM_PCHAN_NONE);
 		break;
 	case GSM_PCHAN_TCH_F_PDCH:
-		ts->pchan_is = GSM_PCHAN_TCH_F;
+		ts_set_pchan_is(ts, GSM_PCHAN_TCH_F);
 		break;
 	default:
-		ts->pchan_is = ts->pchan_on_init;
+		ts_set_pchan_is(ts, ts->pchan_on_init);
 		break;
 	}
-
-	LOG_TS(ts, LOGL_DEBUG, "lchans initialized: %d\n", max_lchans);
 }
 
 static void ts_fsm_not_initialized(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -280,7 +330,7 @@ static void ts_fsm_unused_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 	}
 
 	switch (ts->pchan_on_init) {
-	case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+	case GSM_PCHAN_OSMO_DYN:
 	case GSM_PCHAN_TCH_F_PDCH:
 		if (bts->gprs.mode == BTS_GPRS_NONE) {
 			LOG_TS(ts, LOGL_DEBUG, "GPRS mode is 'none': not activating PDCH.\n");
@@ -375,9 +425,9 @@ static void ts_fsm_wait_pdch_act(struct osmo_fsm_inst *fi, uint32_t event, void
 
 	case TS_EV_PDCH_ACT_NACK:
 		if (ts->pchan_on_init == GSM_PCHAN_TCH_F_PDCH)
-			rate_ctr_inc(&bts_ctrs->ctr[BTS_CTR_RSL_IPA_NACK]);
+			rate_ctr_inc(rate_ctr_group_get_ctr(bts_ctrs, BTS_CTR_RSL_IPA_NACK));
 		else
-			rate_ctr_inc(&bts_ctrs->ctr[BTS_CTR_CHAN_ACT_NACK]);
+			rate_ctr_inc(rate_ctr_group_get_ctr(bts_ctrs, BTS_CTR_CHAN_ACT_NACK));
 		ts->pdch_act_allowed = false;
 		ts_fsm_error(fi, TS_ST_UNUSED, "Received PDCH activation NACK");
 		return;
@@ -417,10 +467,10 @@ static void ts_fsm_pdch_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 
 	/* Set pchan = PDCH status, but double check. */
 	switch (ts->pchan_on_init) {
-	case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+	case GSM_PCHAN_OSMO_DYN:
 	case GSM_PCHAN_TCH_F_PDCH:
 	case GSM_PCHAN_PDCH:
-		ts->pchan_is = GSM_PCHAN_PDCH;
+		ts_set_pchan_is(ts, GSM_PCHAN_PDCH);
 		break;
 	default:
 		ts_fsm_error(fi, TS_ST_BORKEN, "pchan %s is incapable of activating PDCH",
@@ -492,11 +542,11 @@ static void ts_fsm_wait_pdch_deact(struct osmo_fsm_inst *fi, uint32_t event, voi
 	case TS_EV_PDCH_DEACT_ACK:
 		/* Remove pchan = PDCH status, but double check. */
 		switch (ts->pchan_on_init) {
-		case GSM_PCHAN_TCH_F_TCH_H_PDCH:
-			ts->pchan_is = GSM_PCHAN_NONE;
+		case GSM_PCHAN_OSMO_DYN:
+			ts_set_pchan_is(ts, GSM_PCHAN_NONE);
 			break;
 		case GSM_PCHAN_TCH_F_PDCH:
-			ts->pchan_is = GSM_PCHAN_TCH_F;
+			ts_set_pchan_is(ts, GSM_PCHAN_TCH_F);
 			break;
 		default:
 			ts_fsm_error(fi, TS_ST_BORKEN, "pchan %s is incapable of deactivating PDCH",
@@ -512,7 +562,7 @@ static void ts_fsm_wait_pdch_deact(struct osmo_fsm_inst *fi, uint32_t event, voi
 
 	case TS_EV_PDCH_DEACT_NACK:
 		if (ts->pchan_on_init == GSM_PCHAN_TCH_F_PDCH)
-			rate_ctr_inc(&ts->trx->bts->bts_ctrs->ctr[BTS_CTR_RSL_IPA_NACK]);
+			rate_ctr_inc(rate_ctr_group_get_ctr(ts->trx->bts->bts_ctrs, BTS_CTR_RSL_IPA_NACK));
 		/* For Osmocom style dyn TS, there actually is no NACK, since there is no RF Channel
 		 * Release NACK message in RSL. */
 		ts_fsm_error(fi, TS_ST_BORKEN, "Received PDCH deactivation NACK");
@@ -557,7 +607,7 @@ static void ts_fsm_in_use_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 	ts->pdch_act_allowed = true;
 
 	/* For static TS, check validity. For dyn TS, figure out which PCHAN this should become. */
-	ts_for_each_potential_lchan(lchan, ts) {
+	ts_for_n_lchans(lchan, ts, ts->max_lchans_possible) {
 		if (lchan_state_is(lchan, LCHAN_ST_UNUSED))
 			continue;
 
@@ -616,9 +666,9 @@ static void ts_fsm_in_use_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 	}
 
 	/* Make sure dyn TS pchan_is is updated. For TCH/F_PDCH, there are only PDCH or TCH/F modes, but
-	 * for Osmocom style TCH/F_TCH/H_PDCH the pchan_is == NONE until an lchan is activated. */
-	if (ts->pchan_on_init == GSM_PCHAN_TCH_F_TCH_H_PDCH)
-		ts->pchan_is = gsm_pchan_by_lchan_type(activating_type);
+	 * for Osmocom style TCH/F_TCH/H_SDCCH8_PDCH the pchan_is == NONE until an lchan is activated. */
+	if (ts->pchan_on_init == GSM_PCHAN_OSMO_DYN)
+		ts_set_pchan_is(ts, gsm_pchan_by_lchan_type(activating_type));
 	ts_lchans_dispatch(ts, LCHAN_ST_WAIT_TS_READY, LCHAN_EV_TS_READY);
 }
 
@@ -687,8 +737,8 @@ static void ts_fsm_borken_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 	default:
 		ctr = BTS_CTR_TS_BORKEN_FROM_UNKNOWN;
 	}
-	rate_ctr_inc(&bts->bts_ctrs->ctr[ctr]);
-	osmo_stat_item_inc(bts->bts_statg->items[BTS_STAT_TS_BORKEN], 1);
+	rate_ctr_inc(rate_ctr_group_get_ctr(bts->bts_ctrs, ctr));
+	osmo_stat_item_inc(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_TS_BORKEN), 1);
 }
 
 static void ts_fsm_borken(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -712,8 +762,8 @@ static void ts_fsm_borken(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 			struct gsm_bts *bts = ts->trx->bts;
 			/* Late PDCH activation ACK/NACK is not a crime.
 			 * Just process them as normal. */
-			rate_ctr_inc(&bts->bts_ctrs->ctr[BTS_CTR_TS_BORKEN_EV_PDCH_ACT_ACK_NACK]);
-			osmo_stat_item_dec(bts->bts_statg->items[BTS_STAT_TS_BORKEN], 1);
+			rate_ctr_inc(rate_ctr_group_get_ctr(bts->bts_ctrs, BTS_CTR_TS_BORKEN_EV_PDCH_ACT_ACK_NACK));
+			osmo_stat_item_dec(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_TS_BORKEN), 1);
 			ts_fsm_wait_pdch_act(fi, event, data);
 			return;
 		}
@@ -725,8 +775,8 @@ static void ts_fsm_borken(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 			struct gsm_bts *bts = ts->trx->bts;
 			/* Late PDCH deactivation ACK/NACK is also not a crime.
 			 * Just process them as normal. */
-			rate_ctr_inc(&bts->bts_ctrs->ctr[BTS_CTR_TS_BORKEN_EV_PDCH_DEACT_ACK_NACK]);
-			osmo_stat_item_dec(bts->bts_statg->items[BTS_STAT_TS_BORKEN], 1);
+			rate_ctr_inc(rate_ctr_group_get_ctr(bts->bts_ctrs, BTS_CTR_TS_BORKEN_EV_PDCH_DEACT_ACK_NACK));
+			osmo_stat_item_dec(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_TS_BORKEN), 1);
 			ts_fsm_wait_pdch_deact(fi, event, data);
 			return;
 		}
@@ -764,7 +814,8 @@ static void ts_fsm_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *data
 			osmo_fsm_inst_state_chg(fi, TS_ST_NOT_INITIALIZED, 0, 0);
 		OSMO_ASSERT(fi->state == TS_ST_NOT_INITIALIZED);
 		ts_terminate_lchan_fsms(ts);
-		ts->pchan_is = ts->pchan_on_init = GSM_PCHAN_NONE;
+		ts->pchan_on_init = GSM_PCHAN_NONE;
+		ts_set_pchan_is(ts, GSM_PCHAN_NONE);
 		ts_fsm_update_id(ts);
 		break;
 
@@ -773,7 +824,7 @@ static void ts_fsm_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *data
 		if (fi->state != TS_ST_NOT_INITIALIZED)
 			osmo_fsm_inst_state_chg(fi, TS_ST_NOT_INITIALIZED, 0, 0);
 		OSMO_ASSERT(fi->state == TS_ST_NOT_INITIALIZED);
-		ts->pchan_is = GSM_PCHAN_NONE;
+		ts_set_pchan_is(ts, GSM_PCHAN_NONE);
 		ts_lchans_dispatch(ts, -1, LCHAN_EV_TS_ERROR);
 		break;
 
@@ -787,8 +838,8 @@ static void ts_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause ca
 	struct gsm_bts_trx_ts *ts = ts_fi_ts(fi);
 	struct gsm_bts *bts = ts->trx->bts;
 	if (ts->fi->state == TS_ST_BORKEN) {
-		rate_ctr_inc(&bts->bts_ctrs->ctr[BTS_CTR_TS_BORKEN_EV_TEARDOWN]);
-		osmo_stat_item_dec(bts->bts_statg->items[BTS_STAT_TS_BORKEN], 1);
+		rate_ctr_inc(rate_ctr_group_get_ctr(bts->bts_ctrs, BTS_CTR_TS_BORKEN_EV_TEARDOWN));
+		osmo_stat_item_dec(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_TS_BORKEN), 1);
 	}
 }
 
@@ -943,7 +994,7 @@ static struct osmo_fsm ts_fsm = {
 bool ts_is_lchan_waiting_for_pchan(struct gsm_bts_trx_ts *ts, enum gsm_phys_chan_config *target_pchan)
 {
 	struct gsm_lchan *lchan;
-	ts_for_each_potential_lchan(lchan, ts) {
+	ts_for_n_lchans(lchan, ts, ts->max_lchans_possible) {
 		if (lchan->fi->state == LCHAN_ST_WAIT_TS_READY) {
 			if (target_pchan)
 				*target_pchan = gsm_pchan_by_lchan_type(lchan->type);
@@ -984,7 +1035,7 @@ bool ts_is_pchan_switching(struct gsm_bts_trx_ts *ts, enum gsm_phys_chan_config
 		 * waiting for PDCH DEACT (N)ACK */
 		if (target_pchan) {
 			switch (ts->pchan_on_init) {
-			case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+			case GSM_PCHAN_OSMO_DYN:
 				if (target_pchan)
 					*target_pchan = GSM_PCHAN_NONE;
 				break;
